// Copyright 1996-2007 by Jon Dart.  All Rights Reserved.

#include "chess.h"
#include "globals.h"
#include "bearing.h"
#include "movegen.h"
#include "scoring.h"
#include "pinfo.h"
#include "config.h"
#include "bitprobe.h"
#include "tbprobe.h"
#include <stdio.h>

char cmd_buf[CMD_BUF_SIZE];
int cmd_buf_count;

Book *opening_book = NULL;
PositionBook *position_book = NULL;
Log *theLog = NULL;
MoveArray *game_moves;
Options options;
char *programPath;
int EGTBMenCount;
int ncpus = 1;
LockDefine(hash_lock);

#if defined(_MSC_VER) && (_MSC_VER <= 1300)
// workaround for lack of << operator on uint64
// Microsoft support article 168440
extern std::ostream& operator<<(std::ostream& os, uint64 i ) {
    char buf[20];
    sprintf(buf,"%I64d", i );
    os << buf;
    return os;
}
#endif

void init_options(const char *pathName) {
    char rcPath[MAX_PATH];
    strcpy(rcPath,pathName);
#ifdef _WIN32
    char *p = strrchr(rcPath,'\\');
#else
    char *p = strrchr(rcPath,'/');
#endif
    if (p == NULL) 
      strcpy(rcPath,"arasan.rc");
    else
      strcpy(p+1,"arasan.rc");
    // try to read arasan.rc file
    options.init(rcPath);
#ifndef _WIN32
    // Also read .rc from the user's HOME,
    // if there is one.
    if (getenv("HOME")) {
       strcpy(rcPath,getenv("HOME"));
       char *p = strrchr(rcPath,'/');
       if (p == NULL) 
          strcpy(rcPath,"arasan.rc");
       else
          strcpy(p+1,"arasan.rc");
       options.init(rcPath);       
    }
#endif
}

int init_globals(const char *pathName, int open_trace,
int init_bitbases)
{
    // must do this first, many things depend on it:
    Bitmap::init();
    // set up attack bitmaps, etc.
    Bearing::init();
    LockInit(hash_lock);
    programPath = new char[MAX_PATH];
    char tracePath[MAX_PATH];
    strcpy(programPath,pathName);
    strcpy(tracePath,programPath);
    cmd_buf_count = 0;
    *cmd_buf = '\0';
    game_moves = new MoveArray();
    theLog = new Log();
    theLog->clear();
    theLog->write_header();
    Scoring::init();
    Move_Generator::init();
    if (init_bitbases && !initBitbase()) {
#ifdef _WIN32
       ::MessageBox(NULL,"Error: can't start chess engine, missing bitbase files","",MB_OK | MB_ICONINFORMATION);
#else
       cerr << "error: missing bitbase files" << endl;
#endif
       return 0;
    }
    if (open_trace) 
    {
       char *p = strrchr(tracePath,'\\');
       strcpy(p+1,"trace.log");
       freopen(tracePath,"w",stdout);
    }
    return 1;
}

void delayed_init() {
    // If no tablebase directory was specified in the .rc file,
    // try to find the table bases under the arasan install dir
    if (options.search.use_tablebases &&
        options.search.tablebase_location == "") {
    char tbPath[MAX_PATH];
    strcpy(tbPath,programPath);
     const char *dir = "TB";
#ifdef _WIN32
     char *p = strrchr(tbPath,'\\');
#else
     char *p = strrchr(tbPath,'/');
#endif
      if (p)
      {
        // replace it with the tb dir name
	    strcpy(p+1,dir);
      }
      else
         strcpy(tbPath,dir);
      options.search.tablebase_location=tbPath;
    }
    if (options.search.use_tablebases) {
       EGTBMenCount = initTB((char*)options.search.tablebase_location.c_str(),
         options.search.tb_cache_size);
    }
    // also initialize the book here
    if (options.book.book_enabled) {
       opening_book = new Book();
    }
    position_book = new PositionBook();
}

// Make this a __cdecl function so we can pass it to atexit()
void CDECL cleanup_globals()
{
    delete opening_book;
    delete position_book;
    delete game_moves;
    if (theLog) delete theLog;
    delete [] programPath;
    Scoring::dump_stats();
    Scoring::cleanup();
    Move_Generator::cleanup();
    LockDestroy(hash_lock);
    Bitmap::cleanup();
}
