// Copyright 1994-2007 by Jon Dart. All Rights Reserved.

#include "movegen.h"
#include "bearing.h"
#include "attacke.h"
#include "util.h"
#include "debug.h"
#include "moveord.h"
#include "config.h"
#include "bearing.h"
#include "globals.h"
#include <iostream>
#include <fstream>

extern const int Direction[2];

static inline void swap( Move moves[], const int i, const int j){
    Move tmp = moves[j];
    moves[j] = moves[i];
    moves[i] = tmp;
}


void Move_Generator::init() {
}


void Move_Generator::cleanup() {
}


static int IsValidMove(const Board &board, Move killer) {
    PieceType pieceMoved = PieceMoved(killer);
    Square start = StartSquare(killer); Square dest = DestSquare(killer);
    // generally killers are not captures but can be at ply 0
    if (Capture(killer) != EmptyPiece || TypeOfMove(killer) == Promotion)
        return 0;
    if (TypeOfPiece(board[start]) == pieceMoved && PieceColor(board[start]) == board.Side()
    && board[dest] == EmptyPiece) {
        switch(pieceMoved) {
            case Knight:
                return 1;
            case King:
            {
                if (TypeOfMove(killer) == KCastle) {
                    Square kp = board.KingPos(board.Side());
                    CastleType CS = board.CastleStatus(board.Side());
                    return board[kp] == MakePiece(King,board.Side()) &&
                        ((CS == CanCastleEitherSide) ||
                        (CS == CanCastleKSide)) &&
                        board[kp + 1] == EmptyPiece &&
                        board[kp + 2] == EmptyPiece &&
                        board.CheckStatus() != InCheck &&
                        !board.any_attacks(kp + 1,board.OppositeSide()) &&
                        !board.any_attacks(kp + 2,board.OppositeSide());
                }
                else if (TypeOfMove(killer) == QCastle) {
                    Square kp = board.KingPos(board.Side());
                    CastleType CS = board.CastleStatus(board.Side());
                    return board[kp] == MakePiece(King,board.Side()) &&
                        ((CS == CanCastleEitherSide) ||
                        (CS == CanCastleQSide)) &&
                        board[kp - 1] == EmptyPiece &&
                        board[kp - 2] == EmptyPiece &&
                        board[kp - 3] == EmptyPiece &&
                        board.CheckStatus() != InCheck &&
                        !board.any_attacks(kp - 1,board.OppositeSide()) &&
                        !board.any_attacks(kp - 2,board.OppositeSide());
                }
                else
                    return 1;
            }
            case Rook:
            {
                int d = Util::Abs(Bearing::Directions[start][dest]);
                return (d==1 || d==8) && board.Clear(start,dest);
            }
            case Bishop:
            {
                int d = Util::Abs(Bearing::Directions[start][dest]);
                return (d==7 || d==9) && board.Clear(start,dest);
            }
            case Queen:
            {
                int d = Util::Abs(Bearing::Directions[start][dest]);
                return (d==1 || d==8 || d==7 || d==9) && board.Clear(start,dest);
            }
            case Pawn:
            {
                if (board.Side() == White) {
                    if (start-8==dest) return 1;
                    else return (board[start-8] == EmptyPiece && Rank(start,White)==2 && start-16==dest);
                }
                else {
                    if (start+8==dest) return 1;
                    else return (board[start+8] == EmptyPiece && Rank(start,Black)==2 && start+16==dest);
                }
            }
            default:
                break;
        }
    }
    return 0;
}


int Move_Generator::OrderWinningCaptures (int captures, Move hashMove) {
    int index = 0;
    int j = 0;
    int scores[100];
    for (int i = index; i < index+captures; i++) {
        const Square start = StartSquare(moves[i]);
        const PieceType capture = Capture(moves[i]);
        if (MovesEqual(hashMove,moves[i])) {
            SetUsed(moves[i]);
        }
        else if (!IsEmptyPiece(capture)) {
            int start_val;
            start_val = PieceValue(board[start]);
            int capture_val = PieceValue(capture);
            int gain;
            if (capture_val - start_val > 10) {
                gain = capture_val - start_val;
                if (TypeOfMove(moves[i]) == Promotion) gain += (PieceValue(PromoteTo(moves[i])) - PAWN_VALUE);
            }
            else
                gain = attack_estimate(board,moves[i]);
#ifdef _TRACE
            if (master) {
                for (int k=0; k < ply; k++) cout << ' ';
                cout << "move ";
                MoveImage(moves[i],cout);
                cout << " est = " << gain << endl;
            }
#endif
            if (gain >= -10) {
                if (i != j) {
                    swap(moves,i,j);
                }
                scores[j] = gain;
                ++j;
            }
            else {
                losers[losers_count++] = moves[i];
            }
        }
        else if (TypeOfMove(moves[i]) == Promotion) {
            int score = PieceValue(PromoteTo(moves[i])) - PAWN_VALUE;
#ifdef _TRACE
            if (master) {
                for (int k=0; k < ply; k++) cout << ' ';
                cout << "move ";
                MoveImage(moves[i],cout);
                cout << " est = " << score << endl;
            }
#endif
            scores[j] = score;
            swap(moves,i,j);
            ++j;
        }
        else if (TypeOfMove(moves[i]) == EnPassant) {
            int gain = attack_estimate(board,moves[i]);
            if (gain >= 0) {
                scores[j] = gain;
                swap(moves,i,j);
                ++j;
            }
            else
                losers[losers_count++] = moves[i];
        }
    }
    int NumMoves = j-index;
    if (NumMoves > 1) {
        Move_Ordering::sort_moves(moves,scores,NumMoves);
    }
    return NumMoves;
}


Move Move_Generator::NextEvasion() {
    if (index < batch_count)
        return batch[index++];
    switch(phase) {
        case 0:
        {
#ifdef _TRACE
            if (master) {
                indent(ply); cout << "phase 0: ";
            }
#endif
            index = 0;
            batch_count = GenerateEvasionsCaptures(moves);
            if (batch_count) {
                batch_count = OrderWinningCaptures(batch_count,NullMove);
            }
            winners = batch_count;
#ifdef _TRACE
            if (master) {
                cout << " winners=" << batch_count << " losers=" << losers_count << endl;
            }
#endif
            // do pv first
            if (hashMove != NullMove) {
                for (int i = 0; i < batch_count; i++) {
                    if (MovesEqual(hashMove,moves[i])) {
                        if (i!=0) swap(moves,i,0);
                        break;
                    }
                }
            }
            batch = moves;
            ++phase;
            if (index < batch_count) {
                return moves[index++];
            }
        }
        case 1:
        {
            batch_count = GenerateEvasionsNonCaptures(moves);
#ifdef _TRACE
            if (master) {
                indent(ply); cout << "phase 1: non-captures: " << batch_count << endl;
            }
#endif
            // We do not consider an evasion move "forced" if it is a winning
            // capture (even if there is only one singular move), since many
            // checks will be refuted by a simple capture of the checking piece.
            forced = (winners == 0) && (losers_count + batch_count == 1);

            if (forced && batch_count) SetForced(moves[0]);

            batch = moves;
            for (int i = 0; i < batch_count; i++) {
                if (MovesEqual(hashMove,moves[i])) {
                    if (i!=0) swap(moves,i,0);
                    break;
                }
            }
            index = 0;
            ++phase;
            if (index < batch_count) {
                return moves[index++];
            }
        }
        case 2:
        {
#ifdef _TRACE
            if (master) {
                indent(ply); cout << "phase 2: losers: " << losers_count << endl;
            }
#endif
            batch = losers;
            batch_count = losers_count;
            if (forced && batch_count) SetForced(losers[0]);
            index = 0;
            ++phase;
            if (index < batch_count) {
                return batch[index++];
            }
            break;
        }
    }
    return NullMove;
}


int Move_Generator::getBatch(Move *&batch,int &index) {
    unsigned NumMoves  = 0;
    batch = moves;
    for (; NumMoves == 0 && phase<=LOSERS_PHASE; ++phase) {
#ifdef _TRACE
        if (master) {
            indent(ply); cout << "phase = " << phase << endl;
        }
#endif
        switch(phase) {
            case HASH_MOVE_PHASE:
            {
                if (!IsNull(hashMove)) {
                    *moves = hashMove;
                    index = 0;
                    NumMoves = 1;
                }
                break;
            }
            case WINNING_CAPTURE_PHASE:
            {
                int captures = Move_Generator::GenerateCaptures(moves);
                index = 0;
                NumMoves = OrderWinningCaptures(captures,hashMove);
                break;
            }
            case KILLER1_PHASE:
            {
                if (!IsNull(killer1) && !MovesEqual(hashMove,killer1)) {
                    if (IsValidMove(board,killer1)) {
                        moves[NumMoves++] = killer1;
                        index = 0;
                    }
                }
                if (ply>=2 && !IsNull(killer3) && !MovesEqual(hashMove,killer3) &&
                !MovesEqual(killer1,killer3)) {
                    if (IsValidMove(board,killer3)) {
                        moves[NumMoves++] = killer3;
                        index = 0;
                    }
                }
            }
            break;
            case KILLER2_PHASE:
            {
                if (!IsNull(killer2) && !MovesEqual(hashMove,killer2) &&
                (ply<2||!MovesEqual(killer2,killer3))) {
                    if (IsValidMove(board,killer2)) {
                        moves[NumMoves++] = killer2;
                        index = 0;
                    }
                }
                if (ply>=2 && !IsNull(killer4) &&
                    !MovesEqual(hashMove,killer4) &&
                    !MovesEqual(killer1,killer4) &&
                !MovesEqual(killer2,killer4)) {
                    if (IsValidMove(board,killer4)) {
                        moves[NumMoves++] = killer4;
                        index = 0;
                    }
                }
            }
            break;
            case HISTORY_PHASE:
            {
                NumMoves = non_captures_count = GenerateNonCaptures(moves);
                index = 0;
                history_count = 0;
                if (NumMoves) {
                    long max = -1L; int maxi=-1;
                    int maxi2 =-1;
                    int hprune_ok = 1;            //NumMoves > 4;
                    for (unsigned i = 0; i < NumMoves; i++) {
                        if (MovesEqual(hashMove,moves[i])) {
                            SetUsed(moves[i]);
                            continue;
                        }
                        else if (MovesEqual(killer1,moves[i]) ||
                        MovesEqual(killer2,moves[i])) {
                            SetUsed(moves[i]);
                            continue;
                        }
                        else if ((MovesEqual(killer3,moves[i]) ||
                        MovesEqual(killer4,moves[i]))) {
                            SetUsed(moves[i]);
                            continue;
                        }
                        if (context) {
                            int pscore = context->history_score_for_pruning(moves[i],board.Side());
                            if (hprune_ok && pscore < 6*Constants::HISTORY_MAX/10) {
                                SetHPrune(moves[i]);
                            }
                            int score = context->history_score_for_ordering(moves[i],board.Side());
                            if (score >= max) {
                                if (maxi >=0) {
                                    maxi2 = maxi;
                                }
                                maxi = i;
                                max = score;
                            }
                        }
                    }
                    if (maxi != -1) {
                        ClearHPrune(moves[maxi]);
                        swap(moves,history_count++,maxi);
                    }
                    if (maxi2 != -1) {
                        ClearHPrune(moves[maxi2]);
                        swap(moves,history_count++,maxi2);
                    }
                }
                NumMoves = non_captures_count;
                index = 0;
                break;
            }
            case LOSERS_PHASE:
            {
                if (losers_count) {
                    batch = losers;
                }
                NumMoves = losers_count;
                index = 0;
                break;
            }
        }                                         // end switch
    }                                             // end for
#ifdef DEBUG
    for (int = 0; i < NumMoves; i++)
        if (Capture(batch[i])==King) ASSERT(0);
#endif
    return NumMoves;
}


int Move_Generator::GenerateNonCaptures(Move *moves) {
    int NumMoves = 0;
    // castling moves
    const ColorType side = board.Side();
    CastleType CS = board.CastleStatus(side);
    if ((CS == CanCastleEitherSide) ||
    (CS == CanCastleKSide)) {
        const Square kp = board.KingPos(side);
#ifdef _DEBUG
        if (side == White) {
            ASSERT(kp == 60);
            ASSERT(board[kp+3] == whiteRook);
        }
        else {
            ASSERT(kp == 4);
            ASSERT(board[kp+3] == blackRook);
        }
#endif
        if (board[kp + 1] == EmptyPiece &&
            board[kp + 2] == EmptyPiece &&
            board.CheckStatus() == NotInCheck &&
            !board.any_attacks(kp + 1,OppositeColor(side)) &&
            !board.any_attacks(kp + 2,OppositeColor(side)))
            // can castle
            moves[NumMoves++] = CreateMove(kp, kp+2, InvalidPiece,
                King,EmptyPiece, KCastle);
    }
    if ((CS == CanCastleEitherSide) ||
    (CS == CanCastleQSide)) {
        const Square kp = board.KingPos(side);
        if (board[kp - 1] == EmptyPiece &&
            board[kp - 2] == EmptyPiece &&
            board[kp - 3] == EmptyPiece &&
            board.CheckStatus() == NotInCheck  &&
            !board.any_attacks(kp - 1,OppositeColor(side)) &&
            !board.any_attacks(kp - 2,OppositeColor(side)))
            // can castle
            moves[NumMoves++] = CreateMove(kp, kp-2, InvalidPiece,
                King, EmptyPiece, QCastle);
    }
    // non-pawn moves:
    Bitmap pieces(board.occupied[board.Side()]);
    pieces.andNot(board.pawn_bits[board.Side()]);
    Square loc;
    while (pieces.iterate(loc)) {
        switch(TypeOfPiece(board[loc])) {
            case Knight:
            {
                Bitmap dests(Bearing::knight_attacks[loc]);
                dests.andNot(board.all_occupied);
                Square sq;
                while (dests.iterate(sq)) {
                    moves[NumMoves++] =
                        CreateMove(loc,sq,Knight);
                }
                break;
            }
            case King:
            {
                Bitmap dests(Bearing::king_attacks[loc]);
                dests.andNot(board.all_occupied);
                dests.andNot(Bearing::king_attacks[board.KingPos(board.OppositeSide())]);
                Square sq;
                while (dests.iterate(sq)) {
                    moves[NumMoves++] =
                        CreateMove(loc,sq,King);
                }
                break;
            }
            case Bishop:
            {
                const int *data = BishopSquares[loc];
                while (*data != 255 && board[*data] == EmptyPiece) {
                    moves[NumMoves++] =
                        CreateMove(loc,(Square)*data,Bishop);
                    ++data;
                }
                data = BishopSquares[loc] + 8;
                while (*data != 255 && board[*data] == EmptyPiece) {
                    moves[NumMoves++] =
                        CreateMove(loc,(Square)*data,Bishop);
                    ++data;
                }
                data = BishopSquares[loc] + 16;
                while (*data != 255 && board[*data] == EmptyPiece) {
                    moves[NumMoves++] =
                        CreateMove(loc,(Square)*data,Bishop);
                    ++data;
                }
                data = BishopSquares[loc] + 24;
                while (*data != 255 && board[*data] == EmptyPiece) {
                    moves[NumMoves++] =
                        CreateMove(loc,(Square)*data,Bishop);
                    ++data;
                }
                break;
            }
            case Rook:
            {
                const int *data = RookSquares[loc];
                while (*data != 255 && board[*data] == EmptyPiece) {
                    moves[NumMoves++] =
                        CreateMove(loc,(Square)*data,Rook);
                    if (Capture(moves[NumMoves-1])!=EmptyPiece) cout << "error" << endl;
                    ++data;
                }
                data = RookSquares[loc] + 8;
                while (*data != 255 && board[*data] == EmptyPiece) {
                    moves[NumMoves++] =
                        CreateMove(loc,(Square)*data,Rook);
                    ++data;
                }
                data = RookSquares[loc] + 16;
                while (*data != 255 && board[*data] == EmptyPiece) {
                    moves[NumMoves++] =
                        CreateMove(loc,(Square)*data,Rook);
                    ++data;
                }
                data = RookSquares[loc] + 24;
                while (*data != 255 && board[*data] == EmptyPiece) {
                    moves[NumMoves++] =
                        CreateMove(loc,(Square)*data,Rook);
                    ++data;
                }
                break;
            }
            case Queen:
            {
                int j;
                for (j = 0; j < 4; j++) {
                    const int *data = BishopSquares[loc] + (j*8);
                    while (*data != 255 && board[*data] == EmptyPiece) {
                        moves[NumMoves++] =
                            CreateMove(loc,(Square)*data,Queen);
                        ++data;
                    }
                }
                for (j = 0; j < 4; j++) {
                    const int *data = RookSquares[loc] + (j*8);
                    while (*data != 255 && board[*data] == EmptyPiece) {
                        moves[NumMoves++] =
                            CreateMove(loc,(Square)*data,Queen);
                        ++data;
                    }
                }
                break;
            }
        }
    }
    // pawn moves
    if (board.Side() == White) {
        Bitmap pawns(board.pawn_bits[White]);
        pawns.shr8();
        pawns.andNot(board.all_occupied);
        pawns.andNot(Bearing::rank_mask[0]);
        Square sq;
        while (pawns.iterate(sq)) {
            moves[NumMoves++] = CreateMove(sq+8,sq,Pawn);
            if (Rank(sq,White)==3 && board[sq-8] == EmptyPiece)
                moves[NumMoves++] = CreateMove(sq+8,sq-8,Pawn);
        }
    }
    else {
        Bitmap pawns(board.pawn_bits[Black]);
        pawns.shl8();
        pawns.andNot(board.all_occupied);
        pawns.andNot(Bearing::rank_mask[7]);
        Square sq;
        while (pawns.iterate(sq)) {
            moves[NumMoves++] = CreateMove(sq-8,sq,Pawn);
            if (Rank(sq,Black)==3 && board[sq+8] == EmptyPiece)
                moves[NumMoves++] = CreateMove(sq-8,sq+8,Pawn);
        }
    }
    return NumMoves;
}


int Move_Generator::GenerateCaptures(Move * moves) {
    unsigned NumMoves;
    NumMoves = 0;
    Square loc;

    if (board.Side() == White) {
        Bitmap pawns(board.pawn_bits[White]);
        Bitmap pawns1 = pawns;
        pawns1.shr(7);
        pawns1.andNot(Bitmap(MAKELONGLONG(0x01010101,0x01010101)));
        pawns1.And(board.occupied[Black]);
        Square dest;
        while (pawns1.iterate(dest)) {
            Square loc = dest+7;
            int candidate = (Rank(loc,White) == 7);
            if (candidate) {
                moves[NumMoves++] =
                    CreateMove(loc,dest,Queen,Pawn,TypeOfPiece(board[dest]),Promotion);
                moves[NumMoves++] =
                    CreateMove(loc,dest,Knight,Pawn,TypeOfPiece(board[dest]),Promotion);
                if (ply == 0) {
                    moves[NumMoves++] =
                        CreateMove(loc,dest,Rook,Pawn,TypeOfPiece(board[dest]),Promotion);
                    moves[NumMoves++] =
                        CreateMove(loc,dest,Bishop,Pawn,TypeOfPiece(board[dest]),Promotion);
                }
            }
            else
                moves[NumMoves++] =
                    CreateMove(loc,dest,Pawn,TypeOfPiece(board[dest]));
        }
        pawns1 = pawns;
        pawns1.shr(9);
        pawns1.andNot(Bitmap(MAKELONGLONG(0x80808080,0x80808080)));
        pawns1.And(board.occupied[Black]);
        while (pawns1.iterate(dest)) {
            Square loc = dest+9;
            int candidate = (Rank(loc,White) == 7);
            if (candidate) {
                moves[NumMoves++] =
                    CreateMove(loc,dest,Queen,Pawn,TypeOfPiece(board[dest]),Promotion);
                moves[NumMoves++] =
                    CreateMove(loc,dest,Knight,Pawn,TypeOfPiece(board[dest]),Promotion);
                if (ply == 0) {
                    moves[NumMoves++] =
                        CreateMove(loc,dest,Rook,Pawn,TypeOfPiece(board[dest]),Promotion);
                    moves[NumMoves++] =
                        CreateMove(loc,dest,Bishop,Pawn,TypeOfPiece(board[dest]),Promotion);
                }
            }
            else {
                moves[NumMoves++] =
                    CreateMove(loc,dest,Pawn,TypeOfPiece(board[dest]));
            }
        }
        pawns1 = pawns;
        pawns1.And(Bearing::rank_mask[1]);
        pawns1.shr8();
        pawns1.andNot(board.all_occupied);
        while (pawns1.iterate(dest)) {
            Square loc = dest + 8;
            moves[NumMoves++] =
                CreateMove(loc,dest,Queen,Pawn,TypeOfPiece(board[dest]),Promotion);
            moves[NumMoves++] =
                CreateMove(loc,dest,Knight,Pawn,TypeOfPiece(board[dest]),Promotion);
            if (ply == 0) {
                moves[NumMoves++] =
                    CreateMove(loc,dest,Rook,Pawn,TypeOfPiece(board[dest]),Promotion);
                moves[NumMoves++] =
                    CreateMove(loc,dest,Bishop,Pawn,TypeOfPiece(board[dest]),Promotion);
            }
        }
        if (!IsInvalid(board.EnPassantSq())) {
            Square epsq = board.EnPassantSq();
            ASSERT(TypeOfPiece(board[epsq])==Pawn);
            if (File(epsq) != 8 && board[epsq + 1] == whitePawn) {
                Square dest = epsq - 8;
                if (board[dest] == EmptyPiece)
                    moves[NumMoves++] =
                        CreateMove(epsq+1,dest,InvalidPiece,Pawn,Pawn,
                        EnPassant);
            }
            if (File(epsq) != 1 && board[epsq - 1] == whitePawn) {
                Square dest = epsq - 8;
                if (board[dest] == EmptyPiece)
                    moves[NumMoves++] =
                        CreateMove(epsq-1,dest,InvalidPiece,Pawn,Pawn,
                        EnPassant);
            }
        }
    }
    else {
        Bitmap pawns(board.pawn_bits[Black]);
        Bitmap pawns1 = pawns;
        pawns1.shl(7);
        pawns1.andNot(Bitmap(MAKELONGLONG(0x80808080,0x80808080)));
        pawns1.And(board.occupied[White]);
        Square dest;
        while (pawns1.iterate(dest)) {
            Square loc = dest-7;
            int candidate = (Rank(loc,Black) == 7);
            if (candidate) {
                moves[NumMoves++] =
                    CreateMove(loc,dest,Queen,Pawn,TypeOfPiece(board[dest]),Promotion);
                moves[NumMoves++] =
                    CreateMove(loc,dest,Knight,Pawn,TypeOfPiece(board[dest]),Promotion);
                if (ply == 0) {
                    moves[NumMoves++] =
                        CreateMove(loc,dest,Rook,Pawn,TypeOfPiece(board[dest]),Promotion);
                    moves[NumMoves++] =
                        CreateMove(loc,dest,Bishop,Pawn,TypeOfPiece(board[dest]),Promotion);
                }
            }
            else
                moves[NumMoves++] =
                    CreateMove(loc,dest,Pawn,TypeOfPiece(board[dest]));
        }
        pawns1 = pawns;
        pawns1.shl(9);
        pawns1.andNot(Bitmap(MAKELONGLONG(0x01010101,0x01010101)));
        pawns1.And(board.occupied[White]);
        while (pawns1.iterate(dest)) {
            Square loc = dest-9;
            int candidate = (Rank(loc,Black) == 7);
            if (candidate) {
                moves[NumMoves++] =
                    CreateMove(loc,dest,Queen,Pawn,TypeOfPiece(board[dest]),Promotion);
                moves[NumMoves++] =
                    CreateMove(loc,dest,Knight,Pawn,TypeOfPiece(board[dest]),Promotion);
                if (ply == 0) {
                    moves[NumMoves++] =
                        CreateMove(loc,dest,Rook,Pawn,TypeOfPiece(board[dest]),Promotion);
                    moves[NumMoves++] =
                        CreateMove(loc,dest,Bishop,Pawn,TypeOfPiece(board[dest]),Promotion);
                }
            }
            else {
                moves[NumMoves++] =
                    CreateMove(loc,dest,Pawn,TypeOfPiece(board[dest]));
            }
        }
        pawns1 = pawns;
        pawns1.And(Bearing::rank_mask[6]);
        pawns1.shl8();
        pawns1.andNot(board.all_occupied);
        while (pawns1.iterate(dest)) {
            Square loc = dest - 8;
            moves[NumMoves++] =
                CreateMove(loc,dest,Queen,Pawn,TypeOfPiece(board[dest]),Promotion);
            moves[NumMoves++] =
                CreateMove(loc,dest,Knight,Pawn,TypeOfPiece(board[dest]),Promotion);
            if (ply == 0) {
                moves[NumMoves++] =
                    CreateMove(loc,dest,Rook,Pawn,TypeOfPiece(board[dest]),Promotion);
                moves[NumMoves++] =
                    CreateMove(loc,dest,Bishop,Pawn,TypeOfPiece(board[dest]),Promotion);
            }
        }

        if (!IsInvalid(board.EnPassantSq())) {
            Square epsq = board.EnPassantSq();
            ASSERT(TypeOfPiece(board[epsq])==Pawn);
            if (File(epsq) != 8 && board[epsq + 1] == blackPawn) {
                Square dest = epsq + 8;
                if (board[dest] == EmptyPiece)
                    moves[NumMoves++] =
                        CreateMove(epsq+1,dest,InvalidPiece,Pawn,Pawn,
                        EnPassant);
            }
            if (File(epsq) != 1 && board[epsq - 1] == blackPawn) {
                Square dest = epsq + 8;
                if (board[dest] == EmptyPiece)
                    moves[NumMoves++] =
                        CreateMove(epsq-1,dest,InvalidPiece,Pawn,Pawn,
                        EnPassant);
            }
        }
    }

    Bitmap pieces(board.occupied[board.Side()]);
    pieces.andNot(board.pawn_bits[board.Side()]);
    while (pieces.iterate(loc)) {
        const ColorType side = PieceColor(board[loc]);
        switch (TypeOfPiece(board[loc])) {
            case EmptyPiece:
            case Pawn:
                break;
            case Knight:
            {
                Bitmap dests(Bearing::knight_attacks[loc]);
                if (side == White)
                    dests.And(board.occupied[Black]);
                else
                    dests.And(board.occupied[White]);
                Square sq;
                while (dests.iterate(sq)) {
                    moves[NumMoves++] =
                        CreateMove(loc,sq,Knight,TypeOfPiece(board[sq]));
                }
            }
            break;
            case Bishop:
            {
                Square dest;
                Bitmap attacks(DiagAtcksA1(board,loc));
                attacks.Or(DiagAtcksA8(board,loc));
                attacks.And(board.occupied[board.OppositeSide()]);
                while (attacks.iterate(dest)) {
                    moves[NumMoves++] =
                        CreateMove(loc,dest,Bishop,TypeOfPiece(board[dest]));
                }
            }
            break;
            case Rook:
            {
                Square dest;
                Bitmap rank_attacks(RankAttacks(board,loc));
                rank_attacks.Or(FileAttacks(board,loc));
                rank_attacks.And(board.occupied[board.OppositeSide()]);
                while (rank_attacks.iterate(dest)) {
                    moves[NumMoves++] =
                        CreateMove(loc,dest,Rook,TypeOfPiece(board[dest]));
                }
            }
            break;
            case Queen:
            {
                Square dest;
                Bitmap attacks(DiagAtcksA1(board,loc));
                attacks.Or(DiagAtcksA8(board,loc));
                attacks.Or(RankAttacks(board,loc));
                attacks.Or(FileAttacks(board,loc));
                attacks.And(board.occupied[board.OppositeSide()]);
                while (attacks.iterate(dest)) {
                    moves[NumMoves++] =
                        CreateMove(loc,dest,Queen,TypeOfPiece(board[dest]));
                }
            }
            break;
            case King:
            {
                Bitmap dests(Bearing::king_attacks[loc]);
                if (side == White)
                    dests.And(board.occupied[Black]);
                else
                    dests.And(board.occupied[White]);
                dests.andNot(Bearing::king_attacks[board.KingPos(
                    board.OppositeSide())]);
                Square sq;
                while (dests.iterate(sq)) {
                    moves[NumMoves++] =
                        CreateMove(loc,sq,King,TypeOfPiece(board[sq]));
                }
            }
            break;
            case InvalidPiece:
                break;
        }
    }
    moves_generated += NumMoves;
    return NumMoves;
}


Move_Generator::Move_Generator( const Board &ABoard,
                                SearchContext *s, 
                                unsigned curr_ply, Move pvMove,
                                int trace)
:
board(ABoard),
context(s),
ply(curr_ply),
moves_generated(0),
phase(HASH_MOVE_PHASE),
losers_count(0),
index(0),
batch_count(0),
forced(0),
hashMove(pvMove),
master(trace)
{
    killer1 = killer2 = killer3 = killer4 = NullMove;
    if (context) {
        context->get_killers(ply,killer1,killer2);
        if (ply >=2)
            context->get_killers(ply-2,killer3,killer4);
        else
            killer3 = killer4 = NullMove;
    }
}


unsigned Move_Generator::GenerateEvasions(Move * moves, int captures_only) {
    unsigned n = GenerateEvasionsCaptures(moves);
    n += GenerateEvasionsNonCaptures(moves+n);
    return n;
}


unsigned Move_Generator::GenerateEvasionsNonCaptures(Move * moves) {
    int num_moves = 0;
    const Square kp = board.KingPos(board.Side());
    if (num_attacks == 1) {
        // try to interpose a piece
        if (sliding(board[source])) {
            Bitmap btwn_squares;
            board.between(source,kp,btwn_squares);
            if (!btwn_squares.is_clear()) {
                // blocking pawn moves
                if (board.Side() == White) {
                    Bitmap pawns(board.pawn_bits[White]);
                    pawns.shr8();
                    pawns.andNot(board.all_occupied);
                    Bitmap pawns1(pawns);
                    pawns.And(btwn_squares);
                    Square sq;
                    while (pawns.iterate(sq)) {
                        if (!Bearing::is_pinned(board, board.Side(), whitePawn, sq+8, sq)) {
                            if (Rank(sq,board.Side()) == 8) {
                                // interposition is a promotion
                                moves[num_moves++] = CreateMove(board,
                                    sq+8, sq, Queen);
                                moves[num_moves++] = CreateMove(board,
                                    sq+8, sq, Rook);
                                moves[num_moves++] = CreateMove(board,
                                    sq+8, sq, Knight);
                                moves[num_moves++] = CreateMove(board,
                                    sq+8, sq, Bishop);
                            }
                            else {
                                moves[num_moves++] = CreateMove(board, sq+8, sq, InvalidPiece);
                            }
                        }
                    }
                    pawns1.And(Bearing::rank_mask[5]);
                    if (!pawns1.is_clear()) {
                        pawns1.shr8();
                        pawns1.andNot(board.all_occupied);
                        pawns1.And(btwn_squares);
                        while (pawns1.iterate(sq)) {
                            if (!Bearing::is_pinned(board, board.Side(), whitePawn, sq+16, sq))
                                moves[num_moves++] = CreateMove(sq+16,sq,Pawn,EmptyPiece);
                        }
                    }
                }
                else {
                    Bitmap pawns(board.pawn_bits[Black]);
                    pawns.shl8();
                    pawns.andNot(board.all_occupied);
                    Bitmap pawns1(pawns);
                    pawns.And(btwn_squares);
                    Square sq;
                    while (pawns.iterate(sq)) {
                        if (!Bearing::is_pinned(board, board.Side(), blackPawn, sq-8, sq)) {
                            if (Rank(sq,board.Side()) == 8) {
                                // interposition is a promotion
                                moves[num_moves++] = CreateMove(board,
                                    sq-8, sq, Queen);
                                moves[num_moves++] = CreateMove(board,
                                    sq-8, sq, Rook);
                                moves[num_moves++] = CreateMove(board,
                                    sq-8, sq, Knight);
                                moves[num_moves++] = CreateMove(board,
                                    sq-8, sq, Bishop);
                            }
                            else {
                                moves[num_moves++] = CreateMove(board, sq-8, sq, InvalidPiece);
                            }
                        }
                    }
                    pawns1.And(Bearing::rank_mask[2]);
                    if (!pawns1.is_clear()) {
                        pawns1.shl8();
                        pawns1.andNot(board.all_occupied);
                        pawns1.And(btwn_squares);
                        while (pawns1.iterate(sq)) {
                            if (!Bearing::is_pinned(board, board.Side(), blackPawn, sq-16, sq))
                                moves[num_moves++] = CreateMove(sq-16,sq,Pawn,EmptyPiece);
                        }
                    }
                }
                // other blocking pieces
                Bitmap pieces(board.occupied[board.Side()]);
                pieces.andNot(board.pawn_bits[board.Side()]);
                Square loc;
                while (pieces.iterate(loc)) {
                    switch (TypeOfPiece(board[loc])) {
                        case EmptyPiece:
                        case Pawn:
                            break;
                        case Knight:
                        {
                            Bitmap dests(Bearing::knight_attacks[loc]);
                            dests.And(btwn_squares);
                            Square sq;
                            while (dests.iterate(sq)) {
                                if (!Bearing::is_pinned(board, board.Side(), board[loc], loc, sq))
                                    moves[num_moves++] =
                                        CreateMove(loc,sq,Knight,TypeOfPiece(board[sq]));
                            }
                        }
                        break;
                        case Bishop:
                        {
                            Square dest;
                            Bitmap a1(Bearing::diag_a1_mask[loc]);
                            a1.Or(Bearing::diag_a8_mask[loc]);
                            a1.And(btwn_squares);
                            while (a1.iterate(dest)) {
                                if (board.Clear(loc,dest) &&
                                    !Bearing::is_pinned(board, board.Side(), board[loc], loc, dest))
                                    moves[num_moves++] =
                                        CreateMove(loc,dest,Bishop,TypeOfPiece(board[dest]));
                            }
                        }
                        break;
                        case Rook:
                        {
                            Square dest;
                            Bitmap r(Bearing::file_mask[File(loc)-1]);
                            r.Or(Bearing::rank_mask[Rank(loc,Black)-1]);
                            r.And(btwn_squares);
                            while (r.iterate(dest)) {
                                if (board.Clear(loc,dest) && !Bearing::is_pinned(board, board.Side(), board[loc], loc, dest))
                                    moves[num_moves++] =
                                        CreateMove(loc,dest,Rook,TypeOfPiece(board[dest]));
                            }
                        }
                        break;
                        case Queen:
                        {
                            Square dest;
                            Bitmap r(Bearing::file_mask[File(loc)-1]);
                            r.Or(Bearing::rank_mask[Rank(loc,Black)-1]);
                            r.Or(Bearing::diag_a1_mask[loc]);
                            r.Or(Bearing::diag_a8_mask[loc]);
                            r.And(btwn_squares);
                            while (r.iterate(dest)) {
                                if (board.Clear(loc,dest) &&
                                    !Bearing::is_pinned(board, board.Side(), board[loc], loc, dest))
                                    moves[num_moves++] =
                                        CreateMove(loc,dest,Queen,TypeOfPiece(board[dest]));
                            }
                        }
                        break;
                        case King:
                            break;
                        case InvalidPiece:
                            break;
                    }
                }
            }
        }
    }
    // generate evasive moves
    Bitmap b(Bearing::king_attacks[kp]);
    b.andNot(board.occupied[board.Side()]);
    b.andNot(board.all_pawn_attacks(board.OppositeSide()));
    b.andNot(Bearing::king_attacks[board.KingPos(board.OppositeSide())]);
    Square sq;
    while (b.iterate(sq)) {
        if (!board.any_attacks(sq, board.OppositeSide()))
        if (num_attacks > 1 || sq != source) {
            // We need to do some extra checking, since the board
            // info on attacks reflects the state before the move,
            // and we need to be sure that the destination square
            // is not attacked after the move.
            int illegal = 0;
            Bitmap it(king_attacks);
            Square atck_sq;
            while (it.iterate(atck_sq)) {
                Piece attacker = board[(Square)(atck_sq)];
                if (sliding(attacker)) {
                    int dir = Bearing::Directions[atck_sq][sq];
                    // check for movement in the direction of the
                    // attacker:
                    if (dir != 0 && (kp + dir == sq))
                        illegal = 1;
                }
                it.clear(atck_sq);
            }
            if (!illegal) {
                moves[num_moves++] = CreateMove(board, kp, sq, InvalidPiece);
            }
        }
    }
    return num_moves;
}


unsigned Move_Generator::GenerateEvasionsCaptures(Move * moves) {
    int num_moves = 0;
    const Square kp = board.KingPos(board.Side());
    king_attacks = board.calc_attacks(kp, board.OppositeSide());
    ASSERT(!king_attacks.is_clear());
    num_attacks = king_attacks.bit_count();
    if (num_attacks == 1) {
        // try to capture checking piece
        source = (Square)king_attacks.first_one();

        ASSERT(source != InvalidSquare);
        Bitmap atcks(board.calc_attacks(source,board.Side()));
        Square sq;
        while (atcks.iterate(sq)) {
            PieceType capturingPiece = TypeOfPiece(board[sq]);
            if (capturingPiece == King) {
                // We can capture with the king only if the piece
                // checking us is undefended.  But always allow a
                // capture *of* the king - for illegal move detection.
                if (TypeOfPiece(board[source]) == King ||
                !board.any_attacks(source, board.OppositeSide())) {
                    moves[num_moves++] = CreateMove(board, sq, source, InvalidPiece);
                }
            }
            else {
                if (!Bearing::is_pinned(board, board.Side(), board[sq], sq, source)) {
                    if (capturingPiece == Pawn &&
                    Rank(source,board.Side()) == 8) {
                        moves[num_moves++] = CreateMove(board,
                            sq, source, Queen);
                        moves[num_moves++] = CreateMove(board,
                            sq, source, Rook);
                        moves[num_moves++] = CreateMove(board,
                            sq, source, Knight);
                        moves[num_moves++] = CreateMove(board,
                            sq, source, Bishop);
                    }
                    else
                        moves[num_moves++] = CreateMove(board,sq, source, InvalidPiece);
                }
            }
        }
        // Bearing::Attack does not return en passant captures, so try
        // this as a special case
        if (board.EnPassantSq() == source) {
            Square dest = source + RankIncr * Direction[board.Side()];
            Piece myPawn = MakePiece(Pawn,board.Side());
            if (File(source) != 8 && board[source + 1] == myPawn) {
                Piece tmp = board[source];
                Piece &place = (Piece &)board[source];
                place = EmptyPiece;               // imagine me gone
                if (!Bearing::is_pinned(board, board.Side(), board[source+1], source + 1, dest))
                    moves[num_moves++] = CreateMove(source + 1, dest, InvalidPiece, Pawn,
                        Pawn, EnPassant);
                place = tmp;
            }
            if (File(source) != 1 && board[source - 1] == myPawn) {
                Piece tmp = board[source];
                Piece &place = (Piece &)board[source];
                place = EmptyPiece;               // imagine me gone
                if (!Bearing::is_pinned(board, board.Side(), board[source-1], source - 1, dest))
                    moves[num_moves++] = CreateMove(source - 1, dest, InvalidPiece, Pawn,
                        Pawn, EnPassant);
                place = tmp;
            }
        }
    }
    return num_moves;
}


unsigned Move_Generator::GenerateChecks(Move * moves) {
    // Note: doesn't at present generate castling moves that check
    const Square kp = board.KingPos(board.OppositeSide());
    Bitmap pieces(board.occupied[board.Side()]);
    pieces.andNot(board.pawn_bits[board.Side()]);
    Square loc;
    int NumMoves = 0;
    while (pieces.iterate(loc)) {
        switch(TypeOfPiece(board[loc])) {
            case Knight:
            {
                Bitmap dests(Bearing::knight_attacks[loc]);
                dests.andNot(board.all_occupied);
                dests.And(Bearing::knight_attacks[kp]);
                Square sq;
                while (dests.iterate(sq)) {
                    moves[NumMoves++] =
                        CreateMove(loc,sq,Knight,EmptyPiece);
                }
                break;
            }
            case King:
                break;
            case Bishop:
            {
                Bitmap dests(Bearing::diag_a1_mask[kp]);
                dests.Or(Bearing::diag_a8_mask[kp]);
                const int *data = BishopSquares[loc];
                Bitmap b;
                while (*data != 255 && board[*data] == EmptyPiece) {
                    if (dests.is_set(*data) && board.Clear(kp,*data))
                        moves[NumMoves++] =
                            CreateMove(loc,*data,Bishop,EmptyPiece);
                    data++;
                }
                data = BishopSquares[loc] + 8;
                while (*data != 255 && board[*data] == EmptyPiece) {
                    if (dests.is_set(*data) && board.Clear(kp,*data))
                        moves[NumMoves++] =
                            CreateMove(loc,*data,Bishop,EmptyPiece);
                    data++;
                }
                data = BishopSquares[loc] + 16;
                while (*data != 255 && board[*data] == EmptyPiece) {
                    if (dests.is_set(*data) && board.Clear(kp,*data))
                        moves[NumMoves++] =
                            CreateMove(loc,*data,Bishop,EmptyPiece);
                    data++;
                }
                data = BishopSquares[loc] + 24;
                while (*data != 255 && board[*data] == EmptyPiece) {
                    if (dests.is_set(*data) && board.Clear(kp,*data))
                        moves[NumMoves++] =
                            CreateMove(loc,*data,Bishop,EmptyPiece);
                    data++;
                }
                break;
            }
            case Rook:
            {
                Bitmap dests(Bearing::file_mask[File(kp)]);
                dests.Or(Bearing::rank_mask[Rank(kp,Black)]);
                const int *data = RookSquares[loc];
                while (*data != 255 && board[*data] == EmptyPiece) {
                    if (dests.is_set(*data) && board.Clear(kp,*data))
                        moves[NumMoves++] =
                            CreateMove(loc,*data,Rook,EmptyPiece);
                    data++;
                }
                data = RookSquares[loc] + 8;
                while (*data != 255 && board[*data] == EmptyPiece) {
                    if (dests.is_set(*data) && board.Clear(kp,*data))
                        moves[NumMoves++] =
                            CreateMove(loc,*data,Rook,EmptyPiece);
                    data++;
                }
                data = RookSquares[loc] + 16;
                while (*data != 255 && board[*data] == EmptyPiece) {
                    if (dests.is_set(*data) && board.Clear(kp,*data))
                        moves[NumMoves++] =
                            CreateMove(loc,*data,Rook,EmptyPiece);
                    data++;
                }
                data = RookSquares[loc] + 24;
                while (*data != 255 && board[*data] == EmptyPiece) {
                    if (dests.is_set(*data) && board.Clear(kp,*data))
                        moves[NumMoves++] =
                            CreateMove(loc,*data,Rook,EmptyPiece);
                    data++;
                }
                break;
            }
            case Queen:
            {
                int j;
                Bitmap dests(Bitmap::Or(Bearing::file_mask[File(kp)],Bearing::rank_mask[Rank(kp,Black)]));
                dests.Or(Bearing::diag_a1_mask[kp]);
                dests.Or(Bearing::diag_a8_mask[kp]);
                for (j = 0; j < 4; j++) {
                    const int *data = BishopSquares[loc] + (j*8);
                    while (*data != 255 && board[*data] == EmptyPiece) {
                        if (dests.is_set(*data) && board.Clear(kp,*data))
                            moves[NumMoves++] =
                                CreateMove(loc,*data,Queen,EmptyPiece);
                        data++;
                    }
                }
                for (j = 0; j < 4; j++) {
                    const int *data = RookSquares[loc] + (j*8);
                    while (*data != 255 && board[*data] == EmptyPiece) {
                        if (dests.is_set(*data) && board.Clear(kp,*data))
                            moves[NumMoves++] =
                                CreateMove(loc,*data,Queen,EmptyPiece);
                        data++;
                    }
                }
                break;
            }
        }
    }
    // pawn moves
    if (board.Side() == White) {
        Square sq;
        Bitmap pawns(board.pawn_bits[White]);
        pawns.shr8();
        pawns.andNot(board.all_occupied);
        pawns.andNot(Bearing::rank_mask[0]);
        Bitmap pawns1(pawns);
        pawns.And(Bearing::pawn_attacks[kp][White]);
        while (pawns.iterate(sq)) {
            moves[NumMoves++] = CreateMove(sq+8,sq,Pawn,EmptyPiece);
        }
        pawns1.And(Bearing::rank_mask[5]);
        pawns1.shr8();
        pawns1.andNot(board.all_occupied);
        pawns1.And(Bearing::pawn_attacks[kp][White]);
        while (pawns1.iterate(sq)) {
            moves[NumMoves++] = CreateMove(sq+16,sq,Pawn,EmptyPiece);
        }
    }
    else {
        Square sq;
        Bitmap pawns(board.pawn_bits[Black]);
        pawns.shl8();
        pawns.andNot(board.all_occupied);
        pawns.andNot(Bearing::rank_mask[7]);
        Bitmap pawns1(pawns);
        pawns.And(Bearing::pawn_attacks[kp][Black]);
        while (pawns.iterate(sq)) {
            moves[NumMoves++] = CreateMove(sq-8,sq,Pawn,EmptyPiece);
        }
        pawns1.And(Bearing::rank_mask[2]);
        pawns1.shr8();
        pawns1.andNot(board.all_occupied);
        pawns1.And(Bearing::pawn_attacks[kp][Black]);
        while (pawns1.iterate(sq)) {
            moves[NumMoves++] = CreateMove(sq-16,sq,Pawn,EmptyPiece);
        }
    }
    return NumMoves;
}


int Move_Generator::GenerateAllMoves(Move *moves,int repeatable) {
    unsigned NumMoves  = 0;
    if (board.CheckStatus() == InCheck) {
        NumMoves = GenerateEvasions(moves,0);
    }
    else {
        NumMoves += GenerateCaptures(moves+NumMoves);
        NumMoves += GenerateNonCaptures(moves+NumMoves);
    }
    if (repeatable) {
        int scores[Constants::MaxMoves];
        for (unsigned i = 0; i < NumMoves; i++) {
            scores[i] = (int)StartSquare(moves[i]) + (int)(DestSquare(moves[i]) << 7);
            switch (PromoteTo(moves[i])) {
                case Queen:
                    scores[i] &= 0xB000;
                    break;
                case Rook:
                    scores[i] &= 0x8000;
                    break;
                case Bishop:
                    scores[i] &= 0x4000;
                    break;
                default:
                    break;
            }
        }
        Move_Ordering::sort_moves(moves, scores, NumMoves);
    }
    return NumMoves;
}
