// Copyright 1987-2007 by Jon Dart.  All Rights Reserved.

#include "bearing.h"
#include "search.h"
#include "scoring.h"
#include "constant.h"
#include "movegen.h"
#include "moveord.h"
#include "util.h"
#include "pinfo.h"
#include "attacke.h"
#include "log.h"
#include "globals.h"
#include "notation.h"
#include "debug.h"
#include "config.h"
#include "tbprobe.h"
#include "legal.h"

#ifdef _WIN32
static const UINT SEARCH_TIMER = 2;
#include <process.h>
#include <wincon.h>
#endif
#include <iostream>
#include <iomanip>
#include <limits.h>
#include <stdio.h>
#include <ctype.h>
#include <errno.h>

void add_pending(char *buf);

static const int Illegal = -Constants::BIG - 1000;
static int Time_Check_Interval = 5000;

#if defined(EXTENDED_FUTILITY_PRUNE)
const int FUTILITY_RANGE = EXTENDED_FUTILITY_DEPTH;
#else
const int FUTILITY_RANGE = FUTILITY_DEPTH;
#endif

static const int limits[64] =
{
    0, 2,   4,  6,  8, 10, 12, 14, 16, 18,
    20, 21, 22, 23, 24, 25, 26, 27, 28, 29,
    30, 31, 32, 33, 34, 35, 36, 37, 38, 39,
    40, 41, 42, 43, 44, 45, 46, 47, 48, 49,
    50, 51, 52, 53, 54, 55, 56, 57, 58, 59,
    60, 61, 62, 63, 64, 65, 66, 67, 68, 69,
    70, 71, 72, 73
};

static int SPLIT_DEPTH[32];

void SearchController::init() {
    // Allocate the hash table.  Size is in kilobytes.  Guesstimate
    // a conversion to the number of hash table entries.
    int hash_entries;
    use_hash_table = options.search.hash_table_size > 0;
    hash_entries = (1000*options.search.hash_table_size)/sizeof(Hash_Entry);
    if (hash_table_size != options.search.hash_table_size) {
        // hash table was resized (GUI can do this)
        reinit();
    }
    if (use_hash_table) {
        initHash(hash_entries);
        hash_table_size = options.search.hash_table_size;
    }
    for (int i = 0; i < 32; i++ ) {
      if (i<=6) SPLIT_DEPTH[i] = (DEPTH_INCREMENT*THREAD_SPLIT_MIN_DEPTH*7)/4;
      else if (i<=8) SPLIT_DEPTH[i] = (DEPTH_INCREMENT*THREAD_SPLIT_MIN_DEPTH*5)/4;
      else SPLIT_DEPTH[i] = DEPTH_INCREMENT*THREAD_SPLIT_MIN_DEPTH;
    }
}

void SearchController::cleanup(int clearHash)
{
  delete rootSearch;
#ifdef SMP
  delete pool;
#endif
  LockDestroy(input_lock);
}

void SearchController::clearHashTables() {
   rootSearch->clearHashTables();
#ifdef SMP
   pool->clearHashTables();
#endif 
   clearHash(); 
}

long SearchController::calc_hash_size()
{
    // Size the hash table according to how much memory we have.
    static int sizes[12]= {
        997, 1997, 3001, 4003, 4999, 6007, 6997, 8003,
        10007, 12007, 15013, 18103
    };

    int i = 0;
#ifdef _WIN32
    SIZE_T memSize = GlobalCompact(0);
    for (i = 0; i < 11; i++)
        if (sizeof(Position_Info)*((long)sizes[i]) > memSize/16)
            break;
    return 32L*sizes[i];
#else
    return 32L*sizes[i];
#endif
}

SearchController::SearchController()
:
command_function(NULL),post_function(NULL),terminate_function(NULL),
pending_function(NULL),
rootSearch(NULL),
age(0),
ratingDiff(0),
ratingFactor(0)
{
    LockInit(input_lock);
    hash_table_size = options.search.hash_table_size;
#ifdef SMP
    pool = new ThreadPool(this,options.search.ncpus);
    ThreadInfo *ti = pool->mainThread();
    ti->state = ThreadInfo::Working;
    rootSearch = (RootSearch*)ti->work;
    ti->parentNode = ti->work->getNode();
#else
    rootSearch = new RootSearch(this);
#endif

    init();
#ifdef _WIN32
    HANDLE inh = GetStdHandle(STD_INPUT_HANDLE);
    DWORD dw;
    pipe = !GetConsoleMode(inh, &dw);
    if (!pipe) {
        SetConsoleMode(inh, dw & ~(ENABLE_LINE_INPUT |
            ENABLE_MOUSE_INPUT | ENABLE_WINDOW_INPUT));
    }
#endif
}


SearchController::~SearchController()
{
    cleanup(1);
}

Move SearchController::find_best_move(
const Board &board,
Search_Type srcType,
int32 timelimit,
int32 xtratime,
int plylimit,
int background,
int isUCI,
Statistics &stat_buf,
Talk_Level t)
{
    explicit_excludes = 0;
    Move excludes[1];
    return find_best_move(board,srcType,timelimit,xtratime,plylimit,background,
        isUCI, stat_buf, t, excludes, 0);
}

Move SearchController::find_best_move(
const Board &board,
Search_Type srcType,
int32 timelimit,
int32 xtratime,
int plylimit,
int background,
int isUCI,
Statistics &stat_buf,
Talk_Level t,
Move *exclude, int num_exclude)
{
    uci = isUCI;
    if (num_exclude) explicit_excludes = 1;
    // Positions are stored in the hashtable with an "age" to identify
    // which search they came from.  "Newer" positions can replace
    // "older" ones. Update the age here since we are starting a
    // new search.
    age = (age + 1) % 3;
    rootSearch->init(board);
    return rootSearch->ply0_search(board,srcType,timelimit,xtratime,plylimit,
                 background,stat_buf,t,exclude,num_exclude);
}

void SearchController::setRatingDiff(int rdiff)
{
    ratingDiff = rdiff;
    ratingFactor = 0;
    if (ratingDiff < -100) {
        ratingFactor = Util::Max((PAWN_VALUE*(ratingDiff+100))/100,-PAWN_VALUE);
    }
    else if (ratingDiff > 100) {
        ratingFactor = Util::Min((PAWN_VALUE*(ratingDiff-100))/100,PAWN_VALUE);
    }
}

int SearchController::draw_score(const Board & board, ColorType computerSide,
       int root_total_mat) {
    int score = 0;
   
    // if we know the opponent's rating (which will be the case if playing
    // on ICC), factor that into the draw score - a draw against a high-rated
    // opponent is good; a draw against a lower-rated one is bad.
    if (ratingDiff != 0) {
       if (board.Side() == computerSide)
          score += ratingFactor;
       else
          score -= ratingFactor;
    }
    // Penalize the side to move for allowing a draw if it is ahead in
    // material
    int mat = Scoring::material_score(board);
    if (mat > 2*PAWN_VALUE) {
      score -= (mat-2*PAWN_VALUE)/8;
    }
    // A draw is worse if there is a lot of material on the board
    if (root_total_mat > 6000) {
        score -= 16;
    }
    else if (root_total_mat > 4800) {
        score -= 16*(root_total_mat-4800)/1200;
    }
    return score;
}

void SearchController::set_background(int b)
{
    rootSearch->set_background(b);
}

void SearchController::reinit()
{
    // After the search options change, we may need to reallocate
    // the hash table.  Free it here so that next time through init()
    // the reallocation will be done.
    freeHash();
}

#ifdef _WIN32
static int is_move(const Board &b,char *buf)
{
    if (isalpha(*buf) && isdigit(buf[1]) &&
        isalpha(buf[2]) && isdigit(buf[3]))
        return 1;
    else if (strncmp(buf,"o-o",3)==0 || strncmp(buf,"O-O",3)==0 ||
        strncmp(buf,"o-o-o",5)==0 || strncmp(buf,"O-O-O",5) ==0)
        return 1;
    else
        return Notation::Value(b,b.Side(),buf) != NullMove;
}
#endif

int SearchController::check_input(const Board &board)
{
    // two threads should not be checking input at once, at least not
    // under Windows
    Lock(input_lock);
#ifdef _WIN32
    DWORD nchar;
    DWORD dw;
    if (pipe) {
        // Use the GnuChess trick of calling PeekNamedPipe to check
        // for user input.  We need to do this because Winboard does
        // not send ^C to interrupt pondering.
        if (PeekNamedPipe(GetStdHandle(STD_INPUT_HANDLE), NULL, 0,
            NULL, &nchar, NULL)) {
            if (nchar && talkLevel == Trace) cout << "(debug) got console input nchar=" << nchar << endl;
            int nextc;
            BOOL terminate;
            for (unsigned i = 0; i < nchar; i++) {
                nextc=getc(stdin);
//                if (talkLevel == Trace) cout << "(debug) got: " << (char)nextc << endl;
                if (nextc == '\n') {
                    cmd_buf[cmd_buf_count]  = '\0';
                    cmd_buf_count = 0;
                    // call a function (supplied by the calling code) that
                    // sets terminate = 1 if we should terminate.
                    command_function(cmd_buf,terminate);
                    if (terminate) {
                        if (talkLevel == Trace) cout << "(debug) terminating" << endl;
			Unlock(input_lock);
                        return 1;
                    }
                }
                else
                    cmd_buf[cmd_buf_count++] = nextc;
            }
            // Hack alert. Sometimes we get a move with no trailing
            // carriage return.
            if (cmd_buf_count > 0) {
                cmd_buf[cmd_buf_count] = '\0';
                if (is_move(board,cmd_buf) || (strncmp(cmd_buf,"usermove",8) == 0 && is_move(board,cmd_buf+9))) {
                    if (talkLevel == Trace) cout << "(debug) got move w/o <CR>: " << cmd_buf << endl;
                    cmd_buf_count = 0;
                    command_function(cmd_buf,terminate);
                    if (terminate) {
			Unlock(input_lock);
                        return 1;
                    }
                }
            }
        }
    }
    else {
        // We are reading from the console, not connected to Winboard.
        // Not something we usually do, except for testing.
        int retVal = WaitForSingleObject(GetStdHandle(STD_INPUT_HANDLE), 0);
        if (retVal == WAIT_OBJECT_0) {
            INPUT_RECORD recBuf[128];
            ReadConsoleInput(GetStdHandle(STD_INPUT_HANDLE),recBuf,
                128,&dw);
            for (unsigned i = 0; i < dw; i++) {
                if (!recBuf[i].Event.KeyEvent.bKeyDown) continue;
                char nextc=recBuf[i].Event.KeyEvent.uChar.AsciiChar;
                if (nextc == (char)13 /* carriage return */) {
                   putchar('\n');
                   cmd_buf[cmd_buf_count]  = '\0';
                   cmd_buf_count = 0;
                   BOOL terminate;
                   // call a function (supplied by the calling code) that
                   // sets terminate = 1 if we should terminate.
                   command_function(cmd_buf,terminate);
                   if (terminate) {
			Unlock(input_lock);
                       return 1;
                   }
                   break;
               }
               else {
                   putchar(nextc);
                   cmd_buf[cmd_buf_count++] = nextc;
               }
           }
       }
   }
   Unlock(input_lock);
   return 0;
#else
   // UNIX version, uses select like Crafty does
   fd_set readfds;
   struct timeval tv;
   int data;

   FD_ZERO(&readfds);
   FD_SET(fileno(stdin), &readfds);
   tv.tv_sec=0;
   tv.tv_usec=0;
   select(16, &readfds, 0, 0, &tv);
   data=FD_ISSET(fileno(stdin), &readfds);
   if (data == -1)
       cerr << "warning: select returned error. errno=" << errno << endl;
   else if (data) {
       // we have something to read
       //     cout << "(debug) got input, data = " << data << endl;
       *cmd_buf = '\0'; cmd_buf_count = 0;
       char buf[256];
       memset(buf,'\0',256);
       int bytes = read(fileno(stdin),buf,255);
       if (bytes <= 0) {
	   Unlock(input_lock);
           return 0;
       }
       char *p = buf;
       int stop = 0;
       while (*p) {
           char *q = cmd_buf;
           while (*p != '\n' && *p != '\0' && cmd_buf_count < 255) {
               *q++ = *p++;
               ++cmd_buf_count;
           }
           *q = '\0';
           p++;                                      // skip over null or CR
           //cout << "(debug) cmd_buf = " << cmd_buf << endl;
           // call a function (supplied by the calling code) that
           // sets terminate = 1 if we should terminate.
           command_function(cmd_buf,stop);
           cmd_buf_count = 0;
           if (stop) {
	     rootSearch->terminate_now();
           }
       }
       Unlock(input_lock);
       return rootSearch->was_terminated();
   }
   Unlock(input_lock);
   return(0);
#endif
}

RootSearch::RootSearch(SearchController *c)
: Search(c) {
   root = c->rootSearch = this;
   parent = this;
   context.clear_killer();
   context.clear_history();
#ifdef SMP
//   ti = controller->pool->mainThread();
#endif
}

Move RootSearch::ply0_search(const Board &board, 
Search_Type srcType,
int32 timelimit,
int32 xtratime,
int plylimit,
const int background,
Statistics &stat_buf,
Talk_Level t,
Move *exclude, int num_exclude) {

    controller->set_talk_level(t);
    time_target = timelimit;
    xtra_time = xtratime;
    ply_limit = plylimit;
    if (ply_limit > 0)
        ply_limit = Util::Min(ply_limit-1,Constants::MaxPly-1);
    type_of_search = srcType;
    num_excludes = Util::Min(MAX_PV,num_exclude);
    if (num_excludes)
       memcpy(excludes,exclude,sizeof(Move)*num_excludes);
    state = NormalState;
    stats = &stat_buf;
    stats->clear();
    side_to_move = background ? board.OppositeSide() : board.Side();
    node = stack; // top of stack
#ifdef EVAL_STATS
    Scoring::clearStats();
#endif
    null_depth = options.search.null_depth_reduction*DEPTH_INCREMENT;

    bkgrnd_search = background;
    Board board_copy(board);
    initial_board = board_copy;
    init(board_copy);

    node->best = NullMove;
    if (Scoring::is_legal_draw(board)) {
        // If it's a legal draw situation before we even move, then
        // just return a draw score and don't search.
        stats->state = Draw;
        stats->value = draw_score(board);
    }
    else {
        // Generate the ply 0 moves here:
        Move_Generator mg(board,&root->context,0,NullMove,1);
        ply0move_count = mg.GenerateAllMoves(ply0moves,0);
        root->stats->num_moves += ply0move_count;
        // Order the moves.  The ordering process removes any illegal moves
        // at ply 0, so fix the move count if necessary.
        stats->mvtot = ply0move_count =
	  order_ply0_moves(initial_board, NullMove);
        // See if there is an "easy" move (a recapture that appears to
        // gain more material than any other alternative).  If so we
        // may be able to terminate the search early.
        easy = (ply0move_count > 0 &&
            board[DestSquare(ply0moves[0])] != EmptyPiece &&
            game_moves->num_moves() > 0 &&
            (DestSquare(game_moves->move(game_moves->num_moves()-1)) ==
            DestSquare(ply0moves[0])) &&
            (Capture(game_moves->move(game_moves->num_moves()-1)) != EmptyPiece &&
            Capture(ply0moves[0]) != EmptyPiece)
            );
        easy_move = ply0moves[0];
        // Do the actual search
        iterate();
    }
    static const int end_of_game[] = {0, 0, 0, 1, 1, 1, 1};
    stats->end_of_game = end_of_game[(int)stats->state];
    if (!controller->uci && !stats->end_of_game && options.search.can_resign) {
        if (stats->depth >= 2) {
            // Use the display value - i.e. do not use score if
            // we are failing low
            if (stats->display_value <= 8-(int)Constants::BIG) {
                // TBD we can sometimes get -10000 as a score!
                // this is bogus, we should not resign.
                if (stats->display_value != -(int)Constants::BIG)
                    stats->state = Resigns;
            }
            else if (stats->display_value > 200-Constants::BIG &&
                ((100*stats->display_value)/PAWN_VALUE <= options.search.resign_threshold))
                stats->state = Resigns;
            stats->end_of_game = end_of_game[(int)stats->state];
        }
    }

    if (controller->talkLevel == Debug) {
        cout.setf(ios::fixed);
        cout << setprecision(2);
        if (stats->elapsed_time > 0) {
            char buf[128];
            stats->print_nps(buf);
            cout << buf << " nodes/second." << endl;
        }
        cout << (stats->num_nodes-stats->num_qnodes) << " regular nodes, " <<
            stats->num_qnodes << " quiescence nodes." << endl;
        cout << stats->hash_searches << " searches of hash table, " <<
            stats->hash_hits << " successful";
        if (stats->hash_searches != 0)
            cout << " (" <<
                (int)((100.0*(float)stats->hash_hits)/((float)stats->hash_searches)) <<
                " percent).";
        cout << endl;
        cout << stats->hash_inserts << " hash entries inserted, " <<
            stats->hash_replaces << " entries replaced, " <<
            stats->hash_inserts_failed << " inserts failed." << endl;
#ifdef MOVE_ORDER_STATS
        cout << "move ordering: ";
        static const char *labels[] = {"1st","2nd","3rd","4th"};
        for (int i = 0; i < 4; i++) {
	  cout << setprecision(2) << labels[i] << " " <<
                (100.0*stats->move_order[i])/(float)stats->move_order_count << "% " ;
        }
        cout << endl;
#endif
        cout <<  "null cutoffs: " << stats->num_null_cuts << " (" <<
            (int)((100.0*(float)stats->num_null_cuts)/((float)stats->num_nodes)) <<
            "%)" << endl;
        cout << "pruning: " <<
#ifdef FUTILITY_PRUNE
            stats->futility_pruning << " moves (futility)" <<
#endif
#ifdef RAZORING
#ifdef FUTILITY_PRUNE
            ", " <<
#endif
            stats->num_razored << " nodes (razored)" <<
#endif
            endl;
        cout << "extensions: " <<
            stats->check_extensions << " check, " <<
            stats->recap_extensions << " recapture, " <<
            stats->pawn_extensions << " pawn, " << endl;
        cout << "            " <<
            stats->threat_extensions << " threat, " <<
	  stats->forced_extensions << " forced move, " << 
	  stats->rep_extensions << " repetition." << endl;
        cout << stats->tb_probes << " tablebase probes, " <<
            stats->tb_hits << " tablebase hits" << endl;
#if defined(SMP) && defined(SMP_STATS)
        cout << stats->splits << " splits," << 
	  " average thread usage=" << (float)(stats->threads)/(float)stats->samples << endl;
#endif
#ifdef EVAL_STATS
        extern long position_hash_probes, position_hash_hits,
            pawn_hash_probes, pawn_hash_hits,
            king_pawn_hash_probes, king_pawn_hash_hits,
            king_cover_hash_probes, king_cover_hash_hits,
            lazy_cut_chances, lazy_cuts;
        cout << "position hash hit rate = " << 100.0*(float)position_hash_hits/(float)position_hash_probes << "%" << endl;
        cout << "pawn hash hit rate     = " << 100.0*(float)pawn_hash_hits/(float)pawn_hash_probes << "%" << endl;
        cout << "king/pawn hash hit rate= " << 100.0*(float)king_pawn_hash_hits/(float)king_pawn_hash_probes << "%" << endl;
        cout << "king cover hash hit rate= " << 100.0*(float)king_cover_hash_hits/(float)king_cover_hash_probes << "%" << endl;
        if (lazy_cuts > 0L)
            cout <<"lazy cutoff in eval    = " << 100.0*(float)lazy_cuts/(float)lazy_cut_chances << "%" << endl;
#endif
        fflush(stdout);
    }
    return node->best;
}

void RootSearch::set_background(int b) {
    bkgrnd_search = b;
    if (!b) {
        // changing from background to foreground search,
        // update display
        show_status(initial_board,getBest(),0,0,0);
    }
}

void RootSearch::show_status(const Board &board,const Move best,int faillow,
 int failhigh,int complete)
{
    if (was_terminated())
        return;
    int ply = stats->depth;
    static char move_text[20];
    Notation::Image(board, best, move_text);
    char val[20];
    stats->complete = complete;
    Scoring::print_score(stats->display_value,val);
    if (controller->talkLevel == Debug) {
        cout.setf(ios::fixed);
        cout << setprecision(2);
        cout << ply << '\t';
        cout << stats->elapsed_time/100.0 << '\t';
        if (faillow) {
	  cout << " --";
	}
        else if (best != NullMove) {
	   cout << move_text;
           if (failhigh) cout << '!';
	}
        cout << '\t' << val << 
                '\t' << stats->num_nodes << endl;
    }
    // Post during ponder if UCI
    if (!is_background_search() || controller->uci) {
      if (controller->post_function) {
            controller->post_function(*stats);
      }
    }
}


void RootSearch::init(const Board &board) {
    if (board.state.moveCount>=90) {
        clearHash();
    }
    time_up = terminated = 0;
    time_added = 0;
    root_total_mat = board.getMaterial(White).value() + board.getMaterial(Black).value();
    start_time = getTime();
}


int RootSearch::check_time(const Board &board,int ply, int rank,
int iteration_depth)
{
    if (was_terminated()) {
        return 1;
    }

    if (controller->command_function && controller->check_input(board)) {
        terminate_now();
        if (controller->talkLevel == Trace)
            cout << "(debug) terminating, command input received" << endl;
        return 1;
    }
#if defined(SMP) && defined(SMP_STATS)
    stats->samples++; 
    stats->threads += controller->pool->activeCount();
#endif
    time_t current_time = (time_t)(getTime());
    if (type_of_search != Fixed_Depth) {
        // update at least the part of the stats structure that
        // is displayed in analysis mode.
        stats->elapsed_time = current_time - start_time;
        if (current_time - (int)get_start_time() > (int)get_time_limit()) {
            if (controller->talkLevel == Trace) {
                terminate_now();
                cout << "(debug) terminating, time up" << endl;
            }
            return 1;
        }
        if (type_of_search == Fixed_Time) {
            return 0;
        }
        unsigned n = theLog->current();

        if (xtra_time > 0 &&
            type_of_search == Time_Limit &&
            (stats->elapsed_time > time_target/2) &&
        time_added == 0) {
            // see if our score has dropped a lot since our
            // last search
            if (n >= 2 && !(*theLog)[n-2].get_book_info().is_valid() &&
                (stats->value < 5*PAWN_VALUE) &&
                (stats->value > -PAWN_VALUE*5) &&
            (stats->value < (*theLog)[n-2].score() - PAWN_VALUE/3)) {
                // search more time because our score is dropping
                time_added++;
                if (controller->talkLevel == Trace) {
                    cout << "(debug) adding time" << endl;
                }
            }
        }
        // Allow terminating early if this is a recapture, appeared
        // better than everything else at initial sort time, and
        // still looks good.
        if (!time_up && easy && iteration_depth >= 5) {
            // do not do this if this is the first move out of book
            if (n>=2 && !(*theLog)[n-2].get_book_info().is_valid() &&
                time_target >= 300 &&
                // Don't do this if our score is dropping ..
                (stats->value > (*theLog)[n-2].score() - PAWN_VALUE/4) &&
                // require that we have consumed at least 1/3 of the time
                // limit:
                (current_time - get_start_time())*3 > time_target) {
                if (getBest() == easy_move &&  PieceMoved(getBest()) != King) {
                    time_up = 1;
                    if (controller->talkLevel == Trace) {
                        cout << "(debug) terminating, easy move" << endl;
                    }
                    terminate_now();
                    return 1;
                }
            }
        }
    }
    if (controller->uci && (current_time-last_time >= 200)) {
      extern int hashSize;
      cout << "info";
      if (stats->elapsed_time>30) cout << " nps " << 
           (long)((100L*stats->num_nodes)/stats->elapsed_time);
      cout << " nodes " << stats->num_nodes << " hashfull " << (1000*stats->new_hash_inserts)/hashSize << endl;
      last_time = current_time;
    }
    if ( time_up) {
        if (controller->talkLevel == Trace)
	  cout << "(debug) terminating, time up" << endl;
        terminate_now();
    }
    return time_up;
}


Move Search::getBest() const {
  return root->stack->best;
}

int Search::draw_score(const Board &board) {
  return controller->draw_score(board,root->side_to_move,root->root_total_mat);
}

void RootSearch::update_stats(int alpha, int beta, int multipv_count)
{
    time_t end_time = getTime();
    stats->elapsed_time = end_time - get_start_time();
    stats->multipv_count = multipv_count;
    // Note: the search routine counts depth internally starting
    // at 0, but we report it starting at 1 (as most other programs do).
    stats->depth = iteration_depth+1;
    stats->value = display_score;
    // Leave the previous value intact if we have failed low.
    if (!(stats->value == alpha && iteration_depth > 0)) {
        stats->display_value = stats->value;
    }
    stats->state = state;
    Board board_copy(root->initial_board);

    // note: retain previous best line if we do not have one here
    if (IsNull(node->pv[0])) {
#ifdef _TRACE
        cout << "(debug) warning: pv is null\n";
#endif
        return;
    }
    else if (node->pv_length == 0) {
        return;
    }
    node->best = node->pv[0]; // ensure "best" is non-null
    ASSERT(!IsNull(node->best));
    *(stats->best_line_image) = '\0';
    stats->best_line[0] = NullMove;
    int i = 0;
    char *p = stats->best_line_image;
#ifdef _TRACE
	cout << "best line:" << endl;
#endif
    while (i < node->pv_length && !IsNull(node->pv[i])) {
        ASSERT(i<Constants::MaxPly);
        stats->best_line[i] = node->pv[i];
        char move_image[20];
        *move_image = '\0';
#ifdef _TRACE
	MoveImage(node->pv[i],cout); cout << ' ' << (flush);
#endif
        ASSERT(legal_move(board_copy,stats->best_line[i]));
        if (controller->uci) {
	   // UCI format for moves
 	   Notation::UCIMoveImage(stats->best_line[i],move_image);
	}
        else {
	  // use SAN
           Notation::Image(board_copy,stats->best_line[i],move_image);
	}
        int len = (int)strlen(move_image);
        if (p+len+1-stats->best_line_image > 255)
            break;
        strcpy(p,move_image);
        p+= len;
        strcpy(p," ");
        p++;
        board_copy.DoMove(stats->best_line[i]);
        ++i;
        int rep_count;
        if (is_draw(board_copy,rep_count,0)) {
            break;
        }
        if (node->pv_length < 2 && controller->use_hash_table) {
            // get the next move from the hash table, if possible
            // (for pondering)
            Position_Info entry;
            Lock(hash_lock);
            Position_Info::ValueType result =
                searchHash(board_copy,board_copy.HashCode(rep_count),
                0,0,0,
                iteration_depth,entry);
            Unlock(hash_lock);
            if (result != Position_Info::NoHit) {
                Move hashMove = entry.best_move(board_copy);
                if (!IsNull(hashMove)) {
                    stats->best_line[i] = hashMove;
                    *p = '\0';
                    if (controller->uci) {
              	       // UCI format for moves
  	               Notation::UCIMoveImage(hashMove,move_image);
 	            }
                    else {
                       // use SAN
                       Notation::Image(board_copy,hashMove,move_image);
		    }
                    int len = (int)strlen(move_image);
                    if (p+len+1-stats->best_line_image > 255)
                        break;
                    strcpy(p,move_image);
                    p+=len;
                    ++i;
                }
                break;
            }
        }
    }
#ifdef _TRACE
    cout << endl;
#endif
    *p = '\0';

    stats->best_line[i] = NullMove;
#ifdef _TRACE
    cout << "best line image: " << stats->best_line_image << endl;
#endif
}


uint64 SearchController::get_numnodes() const
{
    // return # of nodes visited
    return rootSearch ? rootSearch->stats->num_nodes : 0;
}


void RootSearch::terminate_now()
{
    if (controller->get_talk_level() == Trace) 
     cout << "(debug) terminating search" << endl;
    time_up = terminated = terminate = 1;
    stopAllThreads();
}

void RootSearch::iterate()
{
    // Searching may alter the board, so operate on a copy:
    Board board = initial_board;
    Time_Check_Interval = 5000;

    if (is_background_search())
        computerSide = board.OppositeSide();
    else
        computerSide = board.Side();

    terminate = 0;
    time_check_counter = 0;
    last_time = 0;
        
    ply0first = NullMove;
    int value = Scoring::INVALID_SCORE;
    int tb_hit = 0, tb_pieces = 0;
    if (options.search.use_tablebases) {
        const Material &wMat = board.getMaterial(White);
        const Material &bMat = board.getMaterial(Black);
	int tb_score;
        tb_pieces = wMat.men() + bMat.men();
        if(tb_pieces <= EGTBMenCount) {
            stats->tb_probes++;
            tb_hit = probe_tb(board, tb_score, 0);
            if (tb_hit) {
                value = tb_score;
                stats->tb_hits++;
	    }
            if (tb_hit && controller->talkLevel == Trace)
	      cout << "(debug) tb hit, score=" << value << endl;
        }
    }
    if (value == Scoring::INVALID_SCORE) 
       value = display_score = Scoring::evalu8(board,this);
    stats->display_value = value;
    if (type_of_search == Fixed_Time || type_of_search == Time_Limit) 
	   ply_limit = Constants::MaxPly;

    // Incrementally search the board to greater depths - stop when
    // ply limit, time limit, interrupt, or a terminating condition
    // is reached.
    node->best = NullMove;
    for (iteration_depth = 0;
        iteration_depth <= ply_limit && !terminate && !time_up;
        iteration_depth++) {
        if (!controller->explicit_excludes) num_excludes = 0;
        for (int multipv_count=0; multipv_count < options.search.multipv; multipv_count++) {
           int lo_window, hi_window;
           display_score = value;
           lo_window = Util::Max(-Constants::BIG,value - INITIAL_SEARCH_WINDOW / 2);
           hi_window = Util::Min(value + INITIAL_SEARCH_WINDOW / 2,
               Constants::BIG);
           if (controller->get_talk_level() == Trace && is_background_search()) {
	     cout << "(debug) " << iteration_depth << ". move=";
             MoveImage(node->best,cout); cout << " score=" << node->score << endl;
	   }

#ifdef _TRACE
           cout << "iteration " << iteration_depth << " window = [" <<
               lo_window << "," << hi_window << "]" << endl;
#endif
           int failhigh = 0;
           int faillow = 0;
           int oldvalue = -Constants::BIG;
           Move oldbest = NullMove;
           for (int count = 0; 1; ++count,oldvalue = value,oldbest = getBest()) {
               if (!failhigh && !faillow)
		 ply0first = node->pv[0];
               value = ply0_search(board, lo_window, hi_window,
                   DEPTH_INCREMENT*iteration_depth, 
                   excludes,num_excludes,multipv_count);
               update_stats(lo_window,hi_window,multipv_count);
	       //	       MoveImage(node->best,cout); cout<< " ";
	       //               MoveImage(node->pv[0],cout); cout <<endl;
               if (iteration_depth == 0 && Scoring::is_legal_draw(board)) {
                   if (controller->talkLevel == Trace)
                        cout << "(debug) draw, terminating" << endl;
                   terminate_now();
               }
               // check for forced move, but only at depth 2
               // (so we get a ponder move if possible). But exit immediately
	       // if a tb hit because deeper search will hit the q-search and
	       // the score will be inaccurate. Do not terminate here if a
	       // resign score is returned (search deeper to get an accurate
	       // score).
               else if (ply0move_count == 1 && 
                       (iteration_depth >= 2 || tb_hit) && 
                       (!options.search.can_resign || 
                       ((100*stats->display_value)/PAWN_VALUE > 
                         options.search.resign_threshold))) {
                   if (controller->talkLevel == Trace)
                        cout << "(debug) single legal move, terminating" << endl;
                   terminate_now();
               }
               if (!was_terminated() && (state == Checkmate || state == Stalemate)) {
                   if (controller->talkLevel == Trace)
                       cout << "(debug) terminating due to checkmate or statemate, state="
                           << (int)state << endl;
                   terminate_now();
                   break;
               }
               if (stats->elapsed_time > 100) {
                   long denom = bkgrnd_search ? 200L : 500L;
                   Time_Check_Interval = (int)(
                       (denom*stats->num_nodes)/(36L*stats->elapsed_time));
               }
               if (!terminate) {
                   terminate = check_time(board,0,0,iteration_depth);
                   if (terminate && controller->talkLevel == Trace)
                       cout << "(debug) time up" << endl;
                   if (controller->terminate_function &&
                       controller->terminate_function(*stats)) {
                       if (controller->talkLevel == Trace)
                           cout << "(debug) terminating due to program or user input" << endl;
                       terminate_now();
                   }
               }
               if (count == 0 || ((value != oldvalue) || !MovesEqual(getBest(),oldbest))) {
		 show_status(board,getBest(),value <= lo_window,value >= hi_window,value < hi_window);
	       }
               if (terminate) {
                   break;
               }
               else {
                   if (value < hi_window && value > lo_window) {
                       // score is in the bounds of the search window,
                       // no need to re-search
                       break;
                   }
                   if (value >= hi_window) {
#ifdef _TRACE
                       cout << "ply 0 high cutoff, re-searching ..." << endl;
#endif
                       // high cutoff, must re-search
                       ply0first = node->last_move;     // so we research this move
#ifdef TWO_STAGE_WINDOW
		       if (failhigh) {
#endif
                          hi_window = Constants::BIG;
#ifdef TWO_STAGE_WINDOW
		       } else {
			  // try to expand window only a little
			  hi_window = Util::Min(node->score+2*PAWN_VALUE,Constants::BIG);
		       }
#endif
                       lo_window = Util::Min(value,Constants::BIG-Constants::MaxPly);
                       failhigh++;
                       faillow = 0;
                   }
                   else if (value <= lo_window && !failhigh) {
#ifdef _TRACE
                       cout << "ply 0 low cutoff, re-searching ..." << endl;
                       cout << "value = " << value << endl;
#endif
                       hi_window = lo_window;
#ifdef TWO_STAGE_WINDOW
                       if (faillow) {
#endif
                          lo_window = -Constants::BIG;
#ifdef TWO_STAGE_WINDOW
		       } else {
			  // try to expand window only a little
                          lo_window = Util::Max(node->score-2*PAWN_VALUE,-Constants::BIG);
		       }
#endif
                       failhigh = 0;
                       faillow++;
                   }
                   else {
                       // Use infinite window - next search score should be in
                       // bounds.
                       hi_window = Constants::BIG;
                       lo_window = -Constants::BIG;
                       failhigh = faillow = 0;
                   }
               }
           }
#ifdef _TRACE
           cout << iteration_depth+1 << " ply search result: ";
           MoveImage(node->best,cout);
           cout << " value = ";
	   char buf[20];
	   Scoring::print_score(value,buf);
	   cout << buf << endl;
#endif
           if (options.search.multipv > 1) {
	     excludes[num_excludes++] = node->pv[0];
           }
           // allow early termination on a tablebase position, but not
           // if we have >=6 man tbs in use (because tb set may be
           // incomplete - in that case it is better to allow us to
           // search deeper those nodes that don't produce a tb hit).
           //
           if (tb_hit && tb_pieces<6 && iteration_depth>=3 && !IsNull(node->pv[0])) {
              if (controller->talkLevel == Trace)
                   cout << "(debug) terminating, tablebase hit" << endl;
#ifdef _TRACE
              cout << "terminating, tablebase hit" << endl;
#endif
              terminate_now();
              break;
	   }
           if (value < (int)(iteration_depth*2) - Constants::BIG) {
               // We're either checkmated or we certainly will be, so
               // quit searching.
               if (controller->talkLevel == Trace)
                   cout << "(debug) terminating, low score" << endl;
#ifdef _TRACE
               cout << "terminating, low score" << endl;
#endif
               terminate_now();
               break;
           }
           else if (value >= (Constants::BIG - (int)(iteration_depth*2)
           - 1)) {
               if (iteration_depth>=2) {
                   if (controller->talkLevel == Trace)
                       cout << "(debug) terminating, high score" << endl;
#ifdef _TRACE
                   cout << "terminating, mate score" << endl;
#endif
                   terminate_now();
                   break;
               }
           }
        }
    }
}

int RootSearch::ply0_search(Board & board, int alpha, int beta,
            int depth, Move exclude [], int num_exclude, int multipv_count) {

    // implements alpha/beta search for the top most ply.  We use
    // the negascout algorithm.

    time_check_counter++;
    stats->num_nodes++;

    int in_pv = 1;
    int in_check = 0;
    node->extensions = node->extend = 0;
    int initial_alpha = alpha;

    // At this point we need to know if we are in check or not.
    in_check = (board.CheckStatus() == InCheck);
    if (in_check && ply0move_count == 1)
      SetForced(ply0moves[0]);
    Board_State save_state = board.state;

    int try_score = alpha;
    //
    // re-sort the ply 0 moves.
    // Pv first, then killers. Killers at this ply (ply 0)
    // are moves that have previously exceeded alpha.
    //
    reorder_ply0_moves(board,ply0first);
    ASSERT(num_excludes<MAX_PV);
    for (int i = 0; i < num_exclude; i++) {
       for (int j = 0; j < ply0move_count; j++) {
           if (MovesEqual(ply0moves[j],exclude[i])) {
               SetUsed(ply0moves[j]);
           }
       }
    }

    //
    // Search the next ply
    //
    RootNodeInfo *rootNode = (RootNodeInfo*)node;
    // note: keep previous best move intact in case of faillow
    node->pv[0] = NullMove;
    node->pv_length = 0;
    node->cutoff = 0;
    node->extend = 0;
    node->extensions = 0;
    node->num_try = 0;                              // # of legal moves tried
    node->beta = node->hibound = beta;
    node->alpha = alpha;
    node->score = node->alpha;
    node->best_count = 0;
#ifdef SMP
    node->split = NULL;
    // clear split point stack:
    split_stack_end = split_stack;
#endif
    rootNode->move_count = ply0move_count;
    rootNode->moves = ply0moves;
    int move_index = 0;

    while (move_index < ply0move_count && !rootNode->cutoff && !terminate) {
        Move move  = rootNode->last_move = ply0moves[move_index++];
        stats->mvleft = stats->mvtot-move_index;
        if (controller->uci && stats->elapsed_time) {
           cout << "info currmove ";
           char image[10];
 	   Notation::UCIMoveImage(move,image);
 	   cout << image;
           cout << " currmovenumber " << move_index;
           if (stats->elapsed_time >= 30) {
	     cout << " nps " << (100L*stats->num_nodes)/stats->elapsed_time;
             cout << " nodes " << stats->num_nodes;
	   }
	   extern int hashSize;
	   cout << " hashfull " << (1000*stats->new_hash_inserts)/hashSize;
           cout << endl;
        }
        if (IsMate(move)) continue;               // move leads to mate, skip it
        if (IsUsed(move)) continue;               // move was searched already, skip it
#ifdef _TRACE
	//        cout << "window = [" << rootNode->alpha << ',' << rootNode->hibound << ']' <<
        //    " beta = " << rootNode->beta << endl;
        cout << "trying 0. ";
        MoveImage(move,cout);
        cout << " (" << move_index << "/" << ply0move_count;
        cout << ")" << endl;
#endif
#ifdef SMP
        if (!in_pv && 
            depth >= SPLIT_DEPTH[board.getMaterial(board.Side()).material_level()] &&
            !terminate &&
	    (!options.search.use_tablebases || 
             board.getMaterial(White).men() +
	     board.getMaterial(Black).men() > EGTBMenCount)) {
           // try to search this move in a different thread

	   ThreadInfo *ti;
           if ((ti = controller->pool->checkOut(this,rootNode,0,depth))!=NULL) {
              // a thread is available, split off a new search,
              // copying state so it can run independently:
	      ti->work->init(board,this,rootNode,ti,0);
              // start searching in the new thread:
              ti->start();
#ifdef _THREAD_TRACE
	      cout << "split" << endl;
#endif
#if defined(SMP) && defined(SMP_STATS)
              stats->splits++;
#endif
              continue;
           }
        }
#endif
        node->last_move = move;
        board.DoMove(move);
        rootNode->num_try++;
        // note: no extensions at ply 0
        if (depth-DEPTH_INCREMENT >= 0) {
            try_score = -search(board,-rootNode->hibound, -rootNode->alpha,
                1, depth-DEPTH_INCREMENT);
        }
        else {
            try_score = -quiesce(board, -rootNode->hibound, -rootNode->alpha,
                1, 0);
        }
        if (terminate || rootNode->cutoff) {
            board.UndoMove(move,save_state);
            break;
        }
        // We weeded out illegal moves so should never get one here
        ASSERT(try_score != Illegal);
#ifdef _TRACE
        cout << "0. ";
        MoveImage(move,cout);
        cout << ' ' << try_score;
        if (in_pv) cout << " (pv)";
        cout << endl;
#endif
        if (!in_pv &&
        ((try_score > rootNode->alpha) && (try_score < rootNode->beta))) {
            // We failed to get a cutoff and must re-search
#ifdef _TRACE
            cout << "window = [" << rootNode->alpha << "," << rootNode->beta
                 << "]" << endl;
            cout << "score = " << try_score << " - no cutoff, researching .." << endl;
#endif
            if (depth-DEPTH_INCREMENT >= 0) 
                try_score=-search(board,
                    -rootNode->beta,-rootNode->alpha,1,depth-DEPTH_INCREMENT);
            else
                try_score=-quiesce(board,-rootNode->beta,-rootNode->alpha,1,0);
#ifdef _TRACE
            cout << "0. ";
            MoveImage(move,cout);
            cout << ' ' << try_score;
            cout << endl;
#endif
        }
        board.UndoMove(move,save_state);
        if (terminate || rootNode->cutoff) {
            break;
        }
        // If this move leads to being mated, remove it from the
        // ply 0 move list - there is no use searching it again
        if (try_score <= -Constants::BIG+(int)iteration_depth) {
            for (int i = 0; i < ply0move_count; i++) {
                if (MovesEqual(ply0moves[i],move)) {
#ifdef _TRACE
                    cout << "removing move ";
                    MoveImage(move,cout);
                    cout << endl;
#endif
                    SetMate(ply0moves[i]);
                    continue;
                }
            }
        }
        Lock(rootNode->nodeLock);
        if (try_score > rootNode->alpha) {
            if (update_root_move(board,rootNode,rootNode,move,try_score)) {
	      // beta cutoff
               Unlock(rootNode->nodeLock);
               break;
            }
            if (!in_pv) {
              // pv change after 1st move
              show_status(board,move,0,0,0);
	    }
        }
        rootNode->hibound = rootNode->alpha + 1;    // zero-width window
        Unlock(rootNode->nodeLock);
        in_pv = 0;
    }
#ifdef SMP
    // We may have other threads still searching.
    if (rootNode->cutoff || terminate) {
        // We just need to stop the other threads.
        // We don't care about their results.
#ifdef _THREAD_TRACE
        cout << "waiting, discarding results" << endl;
#endif
        stopAllThreads(rootNode);
    }
    waitForThreadCompletion(rootNode);
#endif
#ifdef _THREAD_TRACE
    cout << "cutoff: " << rootNode->cutoff << " best:";
    MoveImage(rootNode->best,cout); cout << " score:" << rootNode->score << endl;
#endif
    if (rootNode->cutoff)
      return rootNode->score;
    if (in_pv && rootNode->score < initial_alpha) {
        // we have failed low and must reset the window
#ifdef _TRACE
        cout << "ply 0 alpha cutoff" << endl;
#endif
        if (controller->talkLevel == Debug) cout << "(debug) ply 0 alpha cutoff" << endl;
        display_score = rootNode->score;
        return rootNode->score;
    }

    if (rootNode->num_try == 0) {
        // no moves were tried
        if (in_check) {
            if (ply0move_count == 0) {            // mate
                rootNode->score = -(Constants::BIG);
                state = Checkmate;
            }
            else {
                // We generated some moves, and some were legal,
                // but we skipped them all.
                rootNode->score = evalu8(board,alpha,beta,0);
            }
        }
        else {                                    // stalemate
            state = Stalemate;
#ifdef _TRACE
            cout << "stalemate!" << endl;
#endif
            rootNode->score = draw_score(board);
        }
    }
#ifdef MOVE_ORDER_STATS
    if (rootNode->num_try && rootNode->alpha != initial_alpha) {
       stats->move_order_count++;
       if (rootNode->best_count<4) {
           stats->move_order[rootNode->best_count]++;
       }
    }
#endif
    ASSERT(rootNode->score >= -Constants::BIG && rootNode->score <= Constants::BIG);
    return rootNode->score;
}

int RootSearch::update_root_move(const Board &board,
NodeInfo *parentNode, NodeInfo *node,
 Move move, int score) {
    parentNode->best = move;
    display_score = node->score = score;
    // unlike higher plies, allow captures as killers here:
    context.set_killer(node->last_move, 0);
    parentNode->alpha = parentNode->score = score;
    parentNode->best_count = node->num_try-1;
    if (score >= parentNode->beta) {
#ifdef _TRACE
       if (master()) 
         cout << "ply 0 beta cutoff" << endl;
#endif
        if (controller->uci) {
          char buf[20];
	  Scoring::print_score_uci(score,buf);
	  cout << "info score " << buf << " lowerbound" << endl;
        }
        // set pv to this move so it is searched first the next time
        parentNode->pv[0] = move;
        parentNode->pv_length = 1;
        parentNode->cutoff++;
        parentNode->score = score;
        return 1; // signal cutoff
    }
    update_pv(board,parentNode,(node+1),move,0);
    return 0; // no cutoff
}

// stop all threads associated with this search
void Search::stopAllThreads() {
    terminateSearch();
#ifdef SMP
    Lock(split_lock);
    if (controller->get_talk_level() == Trace) 
      cout << "(debug) stopAllThreads: " << ti->index << endl;
    split_t *s = split_stack_end;
    ASSERT(split_stack_end-split_stack<=SPLIT_STACK_MAX_DEPTH);
    while (s != split_stack) {
      --s;
      const int max = s->slaves.size();
      for (int i = 0; i<max; i++) {
	(s->slaves[i])->work->stopAllThreads();
      }
    }   
    Unlock(split_lock);
#endif
}

#ifdef SMP
// initialize a Search instance to prepare it for searching at a split point
void Search::init(const Board &aboard, Search *p, NodeInfo *n, ThreadInfo *ti,int ply) {
    parentNode = n;
    parent = p;

    board = aboard;
    this->ti = ti;
    split_stack_end = split_stack;
    terminate = p->terminate;
    parent = p;
    if (parent == NULL) {
       root = (RootSearch*)this;
    }
    else {
       ASSERT(controller);
       root = controller->rootSearch;
    }
    // copy search history 
    // note: we do not copy all of it just the previous node which we need,
    // and the current node
    if (ply==0) {
      *stack = *(p->stack);
    }
    else {
       memcpy(stack+ply-1,p->stack+ply-1,2*sizeof(NodeInfo));
    }
    // set node ptr to our position in the stack
    node = stack+ply;
    ASSERT(!IsNull(node->last_move));
}

#ifdef SMP
// stop all threads at or below node n
void Search::stopAllThreads(NodeInfo *n) {
    split_t *top = n->split;
	if (top == NULL) return; // no slave threads
    Lock(split_lock);
	split_t *s = split_stack_end;
    // in case of many threads, there may be multiple split points
    // at various depths in the current search
    while (s != top) {
      --s;
      ASSERT(s>=split_stack);
      const int max = s->slaves.size();
      for (int i = 0; i<max; i++) {
        // this call has no parameter because all searching in the
	// associated Search class is to stop
	(s->slaves[i])->work->stopAllThreads();
      }
    }   
    Unlock(split_lock);
}

void Search::waitForThreadCompletion(NodeInfo *n) {
    for (;;) {
       Lock(split_lock);
       if (!n->split || !n->split->slaves.size()) {
		   Unlock(split_lock);
		   break;
       }
       // this node has an active split point
       ti->state = ThreadInfo::Waiting;
#ifdef _THREAD_TRACE
       cout << "waiting - thread " << ti->index << endl;
#endif
       Unlock(split_lock);
       if (ti->wait(500)) break;
       if (controller->get_talk_level() == Trace) 
	      cout << "(debug) waited .5 sec" << endl;
    }
    ti->state = ThreadInfo::Working;
#ifdef _THREAD_TRACE
    cout << "out of wait - thread " << ti->index << endl;
#endif
}
#endif
#endif // SMP

void Search::clearHashTables() {
   memset(pawn_hash_table,0xff,PAWN_HASH_SIZE*sizeof(Pawn_Entry));
   memset(king_pawn_hash_table,0xff,KING_PAWN_HASH_SIZE*sizeof(King_Pawn_Entry));
   memset(king_cover_hash_table,0xff,KING_COVER_HASH_SIZE*sizeof(King_Cover_Entry));
   memset(qsearch_cache,0xff,QSEARCH_CACHE_SIZE*sizeof(QSearchHashEntry));
   for (int i = 0; i < QSEARCH_CACHE_SIZE; i++) {
     qsearch_cache[i].best = NullMove;
     qsearch_cache[i].score = Scoring::INVALID_SCORE;
   }
}

inline void swap( Move moves[], const int i, const int j)
{
   Move tmp = moves[j];
   moves[j] = moves[i]; 
   moves[i] = tmp;
}

// move scoring routine, used only at ply 0
int RootSearch::score(Board &board, const Move &move)
{
   int val = 0;
   Square dest_square = DestSquare(move);
   Piece capture = board[dest_square];
   if (!IsEmptyPiece(capture))
   {
      int est = attack_estimate(board,move);
      if (est > 0)
         val += est;
      else // losing capture
         val += 10 + (est/16);
   }
   else if (TypeOfMove(move) == Promotion) {
       val += PieceValue(PromoteTo(move)) - PieceValue(Pawn);
   }
   Board_State stat = board.state;
   board.DoMove(move);
   if (board.any_attacks(board.KingPos(board.OppositeSide()),board.Side()))
   {
      // move into check, illegal
      val = -Constants::BIG;
   }
   else
   {
      val -= Scoring::positional_score(board,this,-Constants::BIG,Constants::BIG);
#ifdef _TRACE
      cout << "position=" << Scoring::positional_score(board,this,-Constants::BIG,Constants::BIG) << endl;
#endif
      if (board.CheckStatus() == InCheck)
         val += 25;
   }
   board.UndoMove(move,stat);
#ifdef _TRACE
   MoveImage(move,cout); cout << " score = " << val << endl;
#endif
   return val;
}

int RootSearch::order_ply0_moves( Board &board, const Move &best_move)
{
   Scoring::init();
   int scores[Constants::MaxMoves];

   int i;
   for (i = 0; i < ply0move_count; i++)
   {
      if (EqualMove(ply0moves[i],best_move))
         scores[i] = Constants::BIG;
      else
         scores[i] = score(board,ply0moves[i]);
   }
   Move_Ordering::sort_moves(ply0moves, scores, ply0move_count);
   int real_num_moves = ply0move_count;

   easy = 1;
   for (i=ply0move_count; i >=0; --i)
   {
      if (scores[i] == -Constants::BIG)
      {
         --real_num_moves;
      }
      else if (i>0 && scores[0]-100 < scores[i])
	easy = 0;
   }
   return real_num_moves;
}


int
RootSearch::reorder_ply0_moves(const Board & board, const Move & best_move)
{
   if (ply0move_count <= 1)
     return ply0move_count;
   int i; int j = 0;
   for (i = 0; i < ply0move_count; i++)
   {
      ClearUsed(ply0moves[i]);
      if (EqualMove(ply0moves[i],best_move))
      {
#ifdef _TRACE
         cout << "move ";
         MoveImage(ply0moves[i],cout);
         cout << " (pv)"<< endl;
#endif
         swap(ply0moves,i,0); ++j;
      }
   }
   for (i=j; i< ply0move_count;i++) {
      if (EqualMove(ply0moves[i],root->context.Killers1[0])) {
         swap(ply0moves,i,j); j++;
         //break; // causes VC++98 compiler error
      }
   }
   for (i=j; i< ply0move_count;i++) {
      if (EqualMove(ply0moves[i],root->context.Killers2[0])) {
         swap(ply0moves,i,j); 
         break;
      }
   }
   return ply0move_count;
}

Search::Search(SearchController *c) 
  : parent(NULL),parentNode(NULL),controller(c),root(c->rootSearch) {
   node = stack; // top of stack
   parent = this; // true for main thread
   Scoring::init();
   LockInit(split_lock);
#ifdef SMP
   ti = NULL;
#endif
   pawn_hash_table_allocated = malloc(PAWN_HASH_SIZE*sizeof(Pawn_Entry)+256);
   pawn_hash_table = (Pawn_Entry*)ALIGN_POINTER(pawn_hash_table_allocated,128);
   memset(pawn_hash_table,0xff,PAWN_HASH_SIZE*sizeof(Pawn_Entry));
   king_pawn_hash_table_allocated = malloc(KING_PAWN_HASH_SIZE*sizeof(King_Pawn_Entry)+256);
   king_pawn_hash_table = (King_Pawn_Entry*)ALIGN_POINTER(king_pawn_hash_table_allocated,128);
   memset(king_pawn_hash_table,0xff,KING_PAWN_HASH_SIZE*sizeof(King_Pawn_Entry));
   king_cover_hash_table_allocated = malloc(KING_COVER_HASH_SIZE*sizeof(King_Cover_Entry)+256);
   qsearch_cache_allocated = malloc(QSEARCH_CACHE_SIZE*sizeof(QSearchHashEntry)+256);
   king_cover_hash_table = (King_Cover_Entry*)ALIGN_POINTER(king_cover_hash_table_allocated,128);
    memset(king_cover_hash_table,0xff,KING_COVER_HASH_SIZE*sizeof(King_Cover_Entry));
    qsearch_cache = (QSearchHashEntry*)ALIGN_POINTER(qsearch_cache_allocated,128);
    memset(qsearch_cache,0xff,QSEARCH_CACHE_SIZE*sizeof(QSearchHashEntry));
    for (int i = 0; i < QSEARCH_CACHE_SIZE; i++) {
      qsearch_cache[i].best = NullMove;
      qsearch_cache[i].score = Scoring::INVALID_SCORE;
    }
}

void Search::init(const Board &aboard, Search *p) {
    board = aboard;
#ifdef SMP
    split_stack_end = split_stack;
#endif
    parent = p;
    if (parent == NULL) {
       root = (RootSearch*)this;
    }
    else {
       ASSERT(controller);
       root = controller->rootSearch;
    }
}


Search::~Search()
{
   LockDestroy(split_lock);
   free(qsearch_cache_allocated);
   free(pawn_hash_table_allocated);
   free(king_pawn_hash_table_allocated);
   free(king_cover_hash_table_allocated);
}

void RootSearch::update_history(NodeInfo *parentNode, Move best, int cutoff, int depth, ColorType side)
{
    for (int i=0; i<parentNode->num_try; i++) {
      ASSERT(i<Constants::MaxMoves);
      const Move &m = parentNode->done[i];
      if (CanHPrune(m)) {
         const int index = SearchContext::history_index(m,board.Side());
         ASSERT(index<8192);
         SearchContext::HistoryEntry *h = &root->context.History[index];
         if (MovesEqual(m,best)) {
           h->hits++;
           if (cutoff) {
 	      int hist = root->context.History2[MakePiece(PieceMoved(m),side)][DestSquare(m)]++;
              if (hist+1 > Constants::HISTORY_MAX) {
                for (int i = 0; i < 16; i++) {
                  for (int j = 0; j < 64; j++) {
                    root->context.History2[i][j] /= 2;
		  }
       	        }
              }
           }
         }
         h->total++;
         if (h->total > Constants::HISTORY_MAX) {
            h->hits = h->hits/2 + 1;
            h->total = h->total/2 + 1;
         }
       }
    }
}

void Search::update_pv(const Board &board,NodeInfo *node,NodeInfo *fromNode,Move move, int ply) {
#ifdef _TRACE
   if (master()) {
      indent(ply); cout << "update_pv, ply " << ply << endl; 
   }
#endif
    node->pv[ply] = move; 
    if (fromNode->pv_length)
       memcpy(node->pv+ply+1,fromNode->pv+ply+1,
              sizeof(Move)*fromNode->pv_length);
    node->pv_length = fromNode->pv_length+1; 
#ifdef _DEBUG
    Board board_copy(board);
    for (int i = ply; i < node->pv_length+ply; i++) {
#ifdef _TRACE
      if (master()) {
         MoveImage(node->pv[i],cout); cout << " " << (flush);
      }
#endif
       ASSERT(legal_move(board_copy,node->pv[i]));
       board_copy.DoMove(node->pv[i]);
   }
#endif
}


void Search::update_pv(const Board &board, const Move &m, int ply) {
   update_pv(board,node,(node+1),m,ply);
#ifdef _DEBUG
   Board board_copy(board);
#ifdef _TRACE
   if (master()) {
      indent(ply); cout << "update_pv, ply " << ply << endl; 
   }
#endif
    for (int i = ply; i < node->pv_length+ply; i++) {
#ifdef _TRACE
      if (master()) {
         MoveImage(node->pv[i],cout); cout << " " << (flush);
      }
#endif
       ASSERT(legal_move(board_copy,node->pv[i]));
       board_copy.DoMove(node->pv[i]);
   }
#endif
#ifdef _TRACE
   if (master()) cout << endl;
#endif
}

int Search::is_draw( const Board &board, int &rep_count, int ply) {
    // First check for draw by repetition
    int entries = board.state.moveCount-2;
    rep_count = 0;
    // follow rule Crafty uses: 2 repeats if ply<=2, 1 otherwise:
    const int target = (ply<=2) ? 2 : 1;
    if (entries >= 0) {
        for (hash_t *rep_list=board.rep_list_head-3;
            entries>=0;
            rep_list-=2,entries-=2) {
            if (*rep_list == board.HashCode()) {
                rep_count++;
                if (rep_count >= target) {
#ifdef _TRACE
                   if (master()) {
                      indent(ply); cout << "rep count=" << rep_count << endl;
                   }
#endif
                   return 1;
                }
            }
        }
    }
#ifdef _TRACE
    if (rep_count && master()) {
        indent(ply); cout << "rep count=" << rep_count << endl;
    }
#endif
    // check the 50 move rule
    if (board.state.moveCount >= 100) {
       return 1;
    }
    // check for insufficient material and other drawing situations
    return Scoring::material_draw(board) || Scoring::is_theoretical_draw(board);
}


int Search::evalu8(const Board & board,
                   int alpha, int beta, int ply)
{
    // Evaluate a board position.
    int rep_count;
    if (is_draw(board,rep_count,ply))
        return draw_score(board);
    int score = Scoring::material_score(board);
    int pos_score = Scoring::positional_score(board,this,-Constants::BIG,
					      Constants::BIG);
    score += pos_score;
    return score;
}

int Search::quiesce(Board & board,int ply,int depth) {
    // recursive function, implements quiescence search.
    //
    ASSERT(ply < Constants::MaxPly);
    root->stats->num_nodes++; 
    root->stats->num_qnodes++;
    root->time_check_counter++; 
    node->pv_length = 0;
    if (ply >= Constants::MaxPly-1) {
        return evalu8(board,node->alpha,node->beta,ply);
    }
#ifdef _TRACE
    if (master()) {
      indent(ply); cout << "window [" << node->alpha << "," 
        << node->beta << "]" << endl;
    }
#endif
    if (board.CheckStatus((node-1)->last_move)==InCheck) {
      if (Bearing::is_pinned(board,board.OppositeSide(),
			     MakePiece(PieceMoved((node-1)->last_move),board.Side()),
  			     StartSquare((node-1)->last_move),
  			     DestSquare((node-1)->last_move)))
        return -Illegal;
      else {
        return qsearch_check(board,ply,depth);
      }
    }
    else {
       return qsearch_no_check(board,ply,depth);
    }
}

int Search::qsearch_no_check(Board & board, int ply, int depth) {
    int stand_pat_score;
    // Establish a default score.  This score is returned if no
    // captures are generated, or if no captures generate a better
    // score (since we generally can choose whether or not to capture).
    int mat_score = Scoring::material_score(board);
    if (mat_score > node->beta + ROOK_VALUE) {
#ifdef _TRACE
        if (master()) {
           indent(ply); cout << "** CUTOFF ** (material)" << endl;;
        }
#endif
        return mat_score;
    }
    int bitscore;
    QSearchHashEntry *cache_entry  = &qsearch_cache[board.HashCode() %
					    QSEARCH_CACHE_SIZE];
    if (board.getMaterial(White).material_level()==0 &&
        board.getMaterial(Black).material_level()==0 &&
        ((bitscore=Scoring::tryBitbase(board))!=Scoring::INVALID_SCORE)) {
         stand_pat_score = bitscore;
    }
    else {
       stand_pat_score = mat_score + 
         Scoring::positional_score(board,this,node->alpha-mat_score,
                                   node->beta-mat_score,cache_entry);
    }
#ifdef _TRACE
    if (master()) {
       indent(ply);
       cout << "cap score = " << stand_pat_score << endl;
    }
#endif
    if (stand_pat_score > node->alpha) {
        if (stand_pat_score >= node->beta) {
#ifdef _TRACE
            if (master()) {
               indent(ply); cout << "**CUTOFF**" << endl;;
            }
#endif
            return stand_pat_score;
        }
        node->alpha = stand_pat_score;
    }

    node->score = node->alpha;
    Move pv;
    if (cache_entry && cache_entry->hash_code == board.HashCode())
      pv = cache_entry->best;
    else
      pv = NullMove;
    
    int initial_alpha = node->alpha;
    int move_index = 0;
    int try_score;
    node->state = board.state;
    const int swag = QSEARCH_FORWARD_PRUNE_BIAS+stand_pat_score;
    if (!IsNull(pv)) {
#ifdef _TRACE
            if (master()) {
              indent(ply);
              cout << "qsearch cache hit: "; MoveImage(pv,cout);
              cout << endl;
            }
#endif
        if (Capture(pv) == King) {
#ifdef _TRACE
            if (master()) {
              indent(ply);
              cout << "previous move illegal, exiting qsearch" << endl;
            }
#endif
            return -Illegal;
        }
	ASSERT(legal_move(board,pv));
        node->last_move = pv;
        board.DoMove(pv);
        try_score = -quiesce(board, -node->beta, -node->alpha, ply+1, depth-1);
        board.UndoMove(pv,node->state);
        if (try_score != Illegal) {
#ifdef _TRACE
            if (master()) {
              indent(ply);
              cout << ply << ". ";
              MoveImage(pv,cout);
              cout << "(pv) " << try_score << endl;
            }
#endif
            if (try_score > node->alpha) {
                node->score = try_score;
                node->best = pv;
                if (try_score >= node->beta)
		  goto search_end;
                node->alpha = try_score;
                if (node->score >= Constants::BIG-1-ply)
                    goto search_end;                    // mating move found
            }
        }
    }
    {
    Move *moves = node->done;
    Move_Generator mg(board, &root->context, ply, NullMove, master());
    // generate all the capture moves
    int move_count = mg.GenerateCaptures(moves);
    root->stats->num_moves += move_count;
#ifdef _TRACE
    int move_count1 = move_count;
#endif
    move_count = Move_Ordering::order_captures(board, moves, move_count, 
          node->alpha,
          stand_pat_score,ply);
#ifdef _TRACE
    if (master()) {
        indent(ply); cout << move_count1 << " capture moves, ";
       cout << move_count << " after pruning." << endl;
    }
#endif

    while (move_index < move_count) {
        Move move = moves[move_index++];
        if (MovesEqual(move,pv)) continue;
        if (Capture(move) == King) {
#ifdef _TRACE
            if (master()) {
              indent(ply);
              cout << "previous move illegal, exiting qsearch" << endl;
            }
#endif
            cache_entry->hash_code = board.HashCode();
            cache_entry->best = move;
            return -Illegal;
        }
        // We already did forward pruning in order_captures, but if alpha has
        // increased we have an opportunity to do it again.
	if (node->alpha != initial_alpha) {
            int est = Gain(move);
            if (swag + est < node->alpha) {
                continue;
            }
            else {
                est = est - PieceValue(PieceMoved(move));
                if (est <= BISHOP_BONUS && Capture(move) != EmptyPiece) {
                    est = attack_estimate(board,move);
                    if (swag + est < node->alpha) {
                        continue;
                    }
                }
            }
	}
#ifdef _TRACE
        if (master()) {
          indent(ply);
          cout << "trying " << ply << ". ";
          MoveImage(move,cout);
          cout << endl;
        }
#endif
        node->last_move = move;
        board.DoMove(move);
        try_score = -quiesce(board, -node->beta, -node->alpha, ply+1, depth-1);
        board.UndoMove(move,node->state);
        if (try_score != Illegal) {
#ifdef _TRACE
            if (master()) {
              indent(ply);
              cout << ply << ". ";
              MoveImage(move,cout);
              cout << ' ' << try_score << endl;
            }
#endif
            if (try_score > node->alpha) {
                node->score = try_score;
                node->best = move;
                if (try_score >= node->beta)
		  goto search_end;
                node->alpha = try_score;
                if (node->score >= Constants::BIG-1-ply)
                    goto search_end;                    // mating move found
            }
        }
#ifdef _TRACE
        else if (master()) {
            indent(ply);
            cout << ply << ". ";
            MoveImage(move,cout);
            cout << " (illegal)" << endl;
        }
#endif
    }
    if (node->score >= node->beta) {
#ifdef _TRACE
        if (master()) {
           indent(ply);
           cout << "**CUTOFF**" << endl;
        }
#endif
        return node->score;
    }
    if ((node->alpha == initial_alpha) &&
	board.getMaterial(board.Side()).material_level() > 8 &&
	(depth >= 1-options.search.checks_in_qsearch)) {
       // Do checks in qsearch 
       Move_Generator mg(board, &root->context, ply, NullMove, master());
       move_count = mg.GenerateChecks(moves);
       move_index = 0;
#ifdef _TRACE
       if (master()) {
	     if (move_count) {
                     indent(ply);
		     cout << move_count << " checks generated" << endl;;
	     }
       }
#endif
       while (move_index < move_count) {
           Move move = moves[move_index++];
           if (MovesEqual(move,pv)) continue;
#ifdef _TRACE
           if (master()) {
             indent(ply);
             cout << "trying " << ply << ". ";
             MoveImage(move,cout);
             cout << endl;
           }
#endif
           node->last_move = move;
           board.DoMove(move);
           try_score = -quiesce(board, -node->beta, -node->alpha, ply+1, depth-1);
           board.UndoMove(move,node->state);
           if (try_score != Illegal) {
#ifdef _TRACE
               if (master()) {
                 indent(ply);
                 cout << ply << ". ";
                 MoveImage(move,cout);
                 cout << ' ' << try_score << endl;
               }
#endif
               if (try_score > node->alpha) {
                   node->score = try_score;
                   node->best = move;
                   if (try_score >= node->beta)
		     goto search_end;
                   node->alpha = try_score;
                   if (node->score >= Constants::BIG-1-ply)
                       goto search_end;          // mating move found
               }
           }
#ifdef _TRACE
           else if (master()) {
               indent(ply);
               cout << ply << ". ";
               MoveImage(move,cout);
               cout << " (illegal)" << endl;
           }
#endif
       }
    }
    }
search_end:
    if (!IsNull(node->best)) {
      cache_entry->hash_code = board.HashCode();
      cache_entry->best = node->best;
    }
    if (node->alpha != initial_alpha && node->score < node->beta) {
        if (IsNull(node->best))
            node->pv_length = 0;
        else {
            update_pv(board,node->best,ply);
        }
    }
    return node->score;
}

int Search::qsearch_check(Board & board,int ply, int depth) {

    int initial_alpha = node->alpha;
    int try_score;
    QSearchHashEntry *cache_entry = &qsearch_cache[board.HashCode() %
					    QSEARCH_CACHE_SIZE];
    Move pvmove;
    if (cache_entry->hash_code == board.HashCode()) {
      pvmove = cache_entry->best;
    }
    else
      pvmove = NullMove;
    Move_Generator mg(board, &root->context, ply, pvmove, master());

    Move move;
    node->state = board.state;
    node->num_try = 0;
    while ((move = mg.NextEvasion()) != NullMove) {
        if (Capture(move) == King) {
#ifdef _TRACE
            if (master()) {
              indent(ply);
              cout << "previous move illegal, exiting qsearch" << endl;
            }
#endif
            return -Illegal;
        }
        // omit non-capture evasions if we have already found a move
        // that gets us above alpha
        if (node->alpha > initial_alpha && !IsForced(move) &&
            Capture(move)==EmptyPiece && PromoteTo(move)==EmptyPiece) {
#ifdef _TRACE
          if (master()) {
             indent(ply);
             cout << "skipping " << ply << ". ";
             MoveImage(move,cout);
             cout << endl;
          }
#endif
	  continue;
	}
#ifdef _TRACE
        if (master()) {
          indent(ply);
          cout << "trying " << ply << ". ";
          MoveImage(move,cout);
          cout << endl;
        }
#endif
        node->last_move = move;
        board.DoMove(move);
        node->extend = (IsForced(move) && depth>-4);
        try_score = -quiesce(board, -node->beta, -node->alpha, ply+1, depth+node->extend-1);
        board.UndoMove(move,node->state);
        if (try_score != Illegal) {
#ifdef _TRACE
            if (master()) {
              indent(ply);
              cout << ply << ". ";
              MoveImage(move,cout);
              cout << ' ' << try_score << endl;
            }
#endif
            node->num_try++;
            if (try_score > node->alpha) {
                node->score = try_score;
                node->best = move;
                if (try_score >= node->beta)
                    break;
                node->alpha = try_score;
                if (node->score >= Constants::BIG-1-ply)
                    break;                        // mating move found
            }
        }
#ifdef _TRACE
        else if (master()) {
            indent(ply);
            cout << ply << ". ";
            MoveImage(move,cout);
            cout << " (illegal)" << endl;
        }
#endif
    }
    if (!IsNull(node->best)) {
      cache_entry->best = node->best;
      cache_entry->hash_code = board.HashCode();
      cache_entry->score = Scoring::INVALID_SCORE;
    }
    if (node->score >= node->beta) {
#ifdef _TRACE
        if (master()) {
           indent(ply);
           cout << "**CUTOFF**" << endl;
        }
#endif
        return node->score;
    }
    if (node->num_try == 0) {
        return -(Constants::BIG - ply); // mate
    }
    if (node->alpha != initial_alpha && node->score < node->beta) {
        if (IsNull(node->best))
            node->pv_length = 0;
        else {
            update_pv(board,node->best,ply);
        }
    }
    return node->score;
}

static int FORCEINLINE passed_pawn_push(const Board &board, Move move) {
  return PieceMoved(move) == Pawn && Rank(DestSquare(move),board.Side())==7;
}

int Search::pruneOk(const Board &board, int alpha,
int depth, Move move, int &fmax, int &mat_score)
{
    int fmargin;
    if (!passed_pawn_push(board,move)) {
        fmargin = (depth>DEPTH_INCREMENT) ? EXTENDED_FUTILITY_MARGIN : FUTILITY_MARGIN;
        if (mat_score == -Constants::BIG)
	  mat_score = Scoring::evalu8(board,this);
        if (fmargin + mat_score  < alpha) {
            fmax = fmargin + mat_score;
            return 1;
        }
    }
    return 0;
}

#ifdef SMP
void Search::searchSMP(ThreadInfo *ti) {
#ifdef _THREAD_TRACE
  if (ti->parentNode->split)
    cout <<"searchSMP, count = " << ti->parentNode->split->slaves.size() << endl;
#endif
    if (parentNode->cutoff) return;
       
    const int in_check = board.CheckStatus() == InCheck;
    node->state = board.state;
    const Move &move = node->last_move;
    node->best = NullMove;
#ifdef _DEBUG
    node->state = board.state;
#endif
    const int ply = node->split->ply;
    const int depth = node->split->depth;
    int try_score;
	
    // make the move 
    board.DoMove(move);

    // check for legality (unless at ply 0)
    if (ply > 0) {
        if (!in_check && 
             board.any_attacks(board.KingPos(
             board.OppositeSide()),board.Side())) {
             // move is illegal
             return;
         }
    }
    CheckStatusType in_check_after_move = board.CheckStatus();
    if (ply > 0 && in_check_after_move == InCheck && node->extend == 0) {
         node->extensions |= CHECK;
         node->extend += CHECK_EXTENSION;
         root->stats->check_extensions++;
         if (ply > limits[root->iteration_depth]) node->extend /= 2;
         if (node->extend > DEPTH_INCREMENT) node->extend = DEPTH_INCREMENT;
    }
    if (depth+node->extend-DEPTH_INCREMENT >= 0) {
         try_score = -search(board,-node->hibound, -node->alpha,
                               ply+1, depth+node->extend-DEPTH_INCREMENT);
    }
    else {
        try_score = -quiesce(board, -node->hibound, -node->alpha,
            ply+1, 0);
    }
    ASSERT(try_score != Illegal);
    if (terminate || parentNode->cutoff) {
        return;
    }
    // it is possible the parent node's alpha has changed, so compare against
    // that
    if (try_score > parentNode->alpha && try_score < node->beta) {
	// failed to get cutoff, research with non-zero window
        if (depth+node->extend-DEPTH_INCREMENT >= 0) {
            try_score = -search(board,-node->beta, -parentNode->alpha,
                                ply+1, depth+node->extend-DEPTH_INCREMENT);
        }
        else {
            try_score = -quiesce(board, -node->beta, -parentNode->alpha,
                 ply+1, 0);
        }
    }
    if ((node->extensions & HPRUNE) &&
        !terminate && !parentNode->cutoff && try_score >= node->beta &&
        try_score != Illegal) {
        // verify LMR
        if (depth-DEPTH_INCREMENT >= 0) {
             try_score = -search(board, -node->beta, -parentNode->alpha,
                 ply+1, depth-DEPTH_INCREMENT);
        }
        else {
            try_score = -quiesce(board, -node->beta, -parentNode->alpha,
                ply+1, 0);
        }
    }
    /**
#ifdef _THREAD_TRACE
    cout << "SMP search, ply " << ply << ": "; MoveImage(move,cout);
    cout << " score=" << try_score << " alpha=" << alpha << " beta=" <<
              beta << endl;
    cout <<"score=" << try_score << endl; 
#endif
    **/
    if (!terminate && !parentNode->cutoff) {
	ti->parentNode->done[ti->parentNode->num_try++] = move;
        if (try_score > ti->parentNode->alpha) {
             // search produced a new best move or cutoff, update parent node
	     int cutoff;
             Lock(parentNode->nodeLock);
#ifdef _THREAD_TRACE
	     cout << "new best smp, ply= " << ply << ' ';
             MoveImage(move,cout); cout << " score=" << try_score << endl;
#endif
#ifdef _DEBUG
	     // normally we do not have to undo the move, since we will just exit this search,
	     // but if debugging, undo it, so we can verify the PV
	     board.UndoMove(move,node->state);
#endif
             if (ply == 0)
                cutoff = ((RootSearch*)(ti->work->parent))->update_root_move(board,ti->parentNode,node,move,try_score);
             else
                 cutoff = ti->work->parent->update_move(board,ti->parentNode,node,move,try_score,ply,depth);
             Unlock(parentNode->nodeLock);
        }
    }
    // this search is done, make this thread available for
    // other tasks:
#ifdef _THREAD_TRACE
    cout << "out of SearchSMP" << endl;
#endif
}
#endif

int Search::update_move(const Board &board, NodeInfo *parentNode, NodeInfo *node, Move move, int score, int ply, int depth) {
    parentNode->score = score;
    parentNode->best = move;
    if (Capture(move) == EmptyPiece &&
        PromoteTo(move) == InvalidPiece) {
        if (node->score >= parentNode->beta) {
          root->context.set_killer(move, ply);
	}
        // wait to update history until all moves have been searched
    }
    parentNode->best_count = node->num_try-1;
    if (score >= parentNode->beta) {
#ifdef _TRACE
      if (master()) {
        indent(ply); cout << "beta cutoff" << endl;
      }
#endif
        parentNode->cutoff++;
    }
    else {
        parentNode->alpha = score;
        // update pv from slave node to master
       update_pv(board,parentNode,(node+1),move,ply);
    }
    return parentNode->cutoff;
}

// recursive function, implements alpha/beta search.  We use
// the PVS (principal variation search) variant of alpha/beta.
int Search::search(Board & board,int ply,int depth) {

    int in_check, threat, rep_count;
    int mat_score;
    int using_tb;
    int move_count;
    node->best_count = 0;
#ifdef FUTILITY_PRUNE
    int fpruned_moves;
#endif
#ifdef RAZORING
    int razor;
#endif
    if (root->time_check_counter++ >= Time_Check_Interval) {
        root->time_check_counter = 0;
        terminate = root->check_time(board,ply,0,root->iteration_depth);
        if (terminate) {
          if (controller->talkLevel == Trace)
              cout << "(debug) terminating, time up" << endl;
	  root->terminate_now(); // signal all searches to end
	}
    }
    ASSERT(ply < Constants::MaxPly);
    root->stats->num_nodes++;
    if (terminate) {
        return node->alpha;
    }
    else if (ply >= Constants::MaxPly-1) {
        return evalu8(board,node->alpha,node->beta,ply);
    }
    int initial_alpha = node->alpha;

    if (is_draw(board,rep_count,ply)) {
#ifdef _TRACE
        if (master()) {
            indent(ply); cout << "draw!" << endl;
        }
#endif
        int score = draw_score(board);
        if (score > node->alpha && score < node->beta) {
            node->pv[ply] = NullMove;
            node->pv_length = 0;
        }
        return score;
    }
    Move hash_move = NullMove;
    using_tb = 0;
    if (options.search.use_tablebases) {
        const Material &wMat = board.getMaterial(White);
        const Material &bMat = board.getMaterial(Black);
        using_tb = (wMat.men() + bMat.men() <= EGTBMenCount) &&
	  (ply <= 2+Util::Max(1,3*root->iteration_depth/4));
    }
    Position_Info hash_entry;
    Position_Info::ValueType result = Position_Info::NoHit;
    // Note: query the hash table before the tablebases, since TB info may
    // be cached. 
    if (controller->use_hash_table) {
        // Search the hash table to see if we have hit this
        // position before.
        //
        Lock(hash_lock);
        result = searchHash(board,board.HashCode(rep_count),
            node->alpha,node->beta,ply,depth,hash_entry);
        Unlock(hash_lock);
        root->stats->hash_searches++;
#ifdef _DEBUG
        if (result != Position_Info::NoHit) {
            // Found the position in the hash table.
            hash_move = hash_entry.best_move(board);
            if (!IsNull(hash_move) && !legal_move(board,hash_move)) {
                ASSERT(0);
            }
        }
#endif
    }
    if (result != Position_Info::NoHit) {
        root->stats->hash_hits++;
        // if using TBs at this ply, do not pull a non-TB entry out of
        // the hashtable. Also do not allow cut of a full node from a
        // cut hash entry.
	if ((node->flags & VERIFY) || 
             (using_tb && depth != Constants::MaxPly*DEPTH_INCREMENT) ||
	    (node->alpha+1<node->beta && !hash_entry.isFull())) {
	  result = Position_Info::Invalid;
	}
        int value = hash_entry.value();
        switch (result) {
            case Position_Info::Valid:
#ifdef _TRACE
                if (master()) {
                    indent(ply);
                    cout << "hash cutoff, type = E" <<
                        " alpha = " << node->alpha << 
                        " beta = " << node->beta <<
                        " value = " << value << endl;
                }
#endif
                if (value > node->alpha && value < node->beta) {
		    hash_move = hash_entry.best_move(board);
                    if (IsNull(hash_move))
                      node->pv_length=0;
                    else {
                       node->pv[ply] = hash_move;
                       node->pv_length = 1;
		    }
#ifdef _TRACE
                    if (master()) {
                        indent(ply); cout << "best line[ply][ply] = ";
                        MoveImage(hash_move,cout);
                        cout << endl;
                    }
#endif
                }
                return value;
            case Position_Info::UpperBound:
#ifdef _TRACE
                if (master()) {
                    indent(ply);
                    cout << "hash cutoff, type = U" <<
                        " alpha = " << node->alpha << 
                        " beta = " << node->beta <<
                        " value = " << value << endl;
                }
#endif
                return value;                     // cutoff
            case Position_Info::LowerBound:
#ifdef _TRACE
                if (master()) {
                    indent(ply);
                    cout << "hash cutoff, type = L" <<
                        " alpha = " << node->alpha << 
                        " beta = " << node->beta <<
                        " value = " << value << endl;
                }
#endif
                return value;                     // cutoff
            default:
                // Note: hash move & flags may be usable even if score is not usable
                hash_move = hash_entry.best_move(board);
                board.set_check_status(hash_entry.in_check() ?
                    InCheck : NotInCheck);
                break;
        }
    }
    if (using_tb && rep_count==0) {
        int tb_hit = 0, tb_score;
        root->stats->tb_probes++;
        tb_hit = probe_tb(board, tb_score, ply);
        if (tb_hit) {
#ifdef _TRACE
            if (master()) {
                indent(ply); cout << "EGTB hit: score " << tb_score << endl;
            }
#endif
            root->stats->tb_hits++;
            int score = tb_score;
            // insert TB info in hash table.
            if (score <= -Constants::BIG+200) {
                score -= ply;
            }
            else if (score >= Constants::BIG-200) {
                score += ply;
            }
#ifdef _TRACE
            if (master() && tb_score != score) {
                indent(ply); cout << "adjusted score " << score << endl;
            }
#endif
            // Put it in with a large depth so we will not
            // overwrite - this entry is "exact" at all
            // search depths, so effectively its depth is infinite.
            Hash_Entry new_pos(board.HashCode(rep_count),
                Constants::MaxPly*DEPTH_INCREMENT,
                controller->age,
                Position_Info::Valid,
                board.CheckStatus() == InCheck,
                score, 1, NullMove
                );
            Lock(hash_lock);
            storeHash(new_pos,*(root->stats));
            Unlock(hash_lock);
            node->score = tb_score; // unadjusted score
            if (node->score > node->alpha && node->score < node->beta) {
               node->pv[ply] = NullMove;
               node->pv_length = 0;
	    }
            return node->score;
        }
    }
#ifdef _TRACE
    if (master() && !IsNull(hash_move)) {
        indent(ply);
        cout << "hash move = ";
        MoveImage(hash_move,cout);
        cout << endl;
    }
#endif
    // At this point we need to know if we are in check or not.
    in_check = 
       (board.CheckStatus((node-1)->last_move) == InCheck);
#ifdef _DEBUG	
	if (!in_check) {
		if(board.calc_attacks(board.KingPos(board.Side()),board.OppositeSide())!=Bitmap(0)) {
			cout << board << endl;
			MoveImage((node-1)->last_move,cout);
			cout << endl;
			ASSERT(0);
		}
	}
        else {
		if(board.calc_attacks(board.KingPos(board.Side()),board.OppositeSide())==Bitmap(0)) {
			cout << board << endl;
			MoveImage((node-1)->last_move,cout);
			cout << endl;
			ASSERT(0);
		}
	}
#endif
#ifdef _TRACE
    if (master() && in_check) { indent(ply); cout << "in_check=" << in_check << endl;}
#endif
#if defined(FUTILITY_PRUNE) || defined(RAZORING)
    mat_score=-Constants::BIG;
#endif
#if defined(FUTILITY_PRUNE)
    fpruned_moves=0;
#endif
#ifdef RAZORING
    razor = 0;
    if (!in_check && board.getMaterial(board.OppositeSide()).men() >= 3
        &&  depth <= RAZOR_DEPTH && depth > RAZOR_DEPTH-RAZOR_REDUCTION) {

        if (mat_score == -Constants::BIG)
            mat_score = Scoring::material_score(board);
        razor = (mat_score + RAZOR_MARGIN) < node->alpha;
    }
#endif

    move_count = 0;

    // Try to get a fast cutoff using a "null move".  Skip if the side
    // to move is in check or if material balance is low enough that
    // zugzwang is a possibility. Do not do null move if this is an
    // IID search, because it will only help us get a cutoff, not a move.
    // Also avoid null move near the 50-move draw limit.
    threat = 0;
    if (root->null_depth && !in_check && (node->beta==node->alpha+1) && 
        !(node->flags & (IID|VERIFY)) &&
        !IsNull((node-1)->last_move) &&
        board.state.moveCount <= 98 &&
        !board.getMaterial(board.OppositeSide()).no_pieces() && 
	board.getMaterial(board.Side()).pieceCount() >= 1) {

        node->last_move = NullMove;
        node->state = board.state;
        board.DoNull();
#ifdef _TRACE
        if (master()) {
            indent(ply);
            cout << "trying " << ply << ". " << "(null)" << endl;
        }
#endif
        int nu_depth = depth-root->null_depth-DEPTH_INCREMENT;
#ifdef ADAPTIVE_NULL
	if (depth >= 9*DEPTH_INCREMENT ||
            (depth >= 6*DEPTH_INCREMENT &&
            board.getMaterial(board.Side()).pieceCount() >= 2))
            nu_depth -= DEPTH_INCREMENT;          // R=3
#endif
        int nscore;
        if (nu_depth >= 0)
	  nscore = -search(board, -node->beta, -node->alpha,
                ply+1, nu_depth);
        else
	  nscore = -quiesce(board, -node->beta, -node->alpha,
                ply+1, 0);
#ifdef _TRACE
        if (master()) {
            indent(ply);
            cout << ply << ". " << "(null)" << ' ' << nscore << endl;
        }
#endif
        board.UndoNull(node->state);
        if (terminate) {
	    node->score = node->alpha;
            goto search_end2;
        }
        else if (nscore >= node->beta) {                // cutoff
#ifdef VERIFY_NULL_SEARCH
	  if (board.getMaterial(board.Side()).pieceCount() <= 3) {
                // in endgame, verify null cutoff with reduced-depth search
                // (idea from Dieter Buerssner)
                //nu_depth = depth-2*DEPTH_INCREMENT;
		if (nu_depth >= 0)
                    nscore = search(board, node->alpha, node->beta,
                        ply+1, nu_depth, VERIFY);
		else
		    nscore = quiesce(board, node->alpha, node->beta,
                        ply+1, 0);
#ifdef _TRACE
		indent(ply); 
		if (nscore>=node->beta) 
                  cout << "null cutoff verified";
                else
		  cout << "null cutoff not verified";
		cout << endl;
#endif
            }
            if (nscore >= node->beta)                   // null cutoff
#endif
            {
#ifdef _TRACE
                if (master()) {
                    indent(ply);
                    cout << "**CUTOFF**" << endl;
                }
#endif
                root->stats->num_null_cuts++;
                node->score = nscore;
                goto search_end;
            }
        }
        if (nscore <= 200-Constants::BIG) {
	  threat++;
        }
    }

    // Use "internal iterative deepening" to get an initial move to try if
    // there is no hash move .. an idea from Crafty (previously used by
    // Hitech).
    if (IsNull(hash_move) && depth > 2*DEPTH_INCREMENT) {
        int d;
	d = Util::Min(depth-2*DEPTH_INCREMENT,depth/2);
#ifdef _TRACE
        if (master()) {
            indent(ply); cout << "== deepening, depth = " << d
				<< endl;
        }
#endif
        // call search routine at lower depth to get a 1st move to try.
	// Note: we do not push down the node stack because we want this
	// search to have all the same parameters (including ply) as the
	// current search, just reduced depth + the IID flag set.
	node->flags |= IID;
        int alpha = node->alpha;
	node->num_try = 0; 
        search(board,ply,d);
	node->flags &= ~IID;
        // reset key params
	node->num_try = 0; 
        node->cutoff = 0;
        node->alpha = node->score = alpha;
        // set hash move to IID search result (may still be null)
        hash_move = node->best;
        // do not retain any pv information from the IID search
        // (can screw up non-IID pv).
        (node+1)->pv[ply+1] = NullMove;
        (node+1)->pv_length = 0;
        if (terminate) {
	    node->score = node->alpha;
 	    return node->alpha;
        }
#ifdef _TRACE
        if (master()) { 
           indent(ply); cout << "== deepening done.";
        }
#endif
#ifdef _DEBUG
        if (!IsNull(hash_move)) {
          ASSERT(legal_move(board,hash_move));
        }
#endif
#ifdef _TRACE
        if (master()) {
            if (!IsNull(hash_move)) {
                indent(ply); cout << "  hash_move = ";
                MoveImage(hash_move,cout);
            }
            cout << endl;
        }
#endif
    }
    {
        Move_Generator mg(board, &root->context, ply, hash_move, master());
        node->state = board.state;
        int try_score;
        // place to taper off extensions:
        const int limit = ply > limits[root->iteration_depth];
        //
        // Now we are ready to loop through the moves from this position
        //
        int first = 1;
        while (!node->cutoff && !terminate) {
            Move move = in_check ? mg.NextEvasion() : mg.NextMove();
            if (IsNull(move)) break;
            if (IsUsed(move)) continue;
            move_count++;
            root->stats->num_moves++;
#ifdef _TRACE
            if (master()) {
                indent(ply);
                cout << "trying " << ply << ". ";
                MoveImage(move,cout);
                if (first) cout << "(pv)";
                cout << " [" << node->alpha << "," << node->beta << "]";
                cout << " d:" << depth << endl;
            }
#endif
            node->last_move = move;
            CheckStatusType in_check_after_move = board.wouldCheck(move);
            node->extend = 0;
            // see if we should apply any extensions at this node.
            node->extensions = 0;
            if (options.search.forced_extensions && IsForced(move)) {
                // one single reply to check
                node->extend += FORCED_EXTENSION;
                node->extensions |= FORCED;
                root->stats->forced_extensions++;
            }
            else if (options.search.check_extensions && in_check_after_move == InCheck) {
                node->extensions |= CHECK;
                node->extend += CHECK_EXTENSION;
                root->stats->check_extensions++;
            }
            else if (options.search.pawn_push_extensions && 
                PieceMoved(move) == Pawn &&
                Rank(DestSquare(move),board.Side())==7) {
                 node->extensions |= PAWN_PUSH;
                 node->extend += PAWN_PUSH_EXTENSION;
                 root->stats->pawn_extensions++;
            }
	    else if (options.search.recapture_extensions &&
                Capture(move) != EmptyPiece &&
                Capture(move) != Pawn &&
	        Capture((node-1)->last_move) != EmptyPiece &&
                DestSquare(move) == DestSquare((node-1)->last_move) &&
		     Util::Abs(PieceValue(Capture((node-1)->last_move))-
			  PieceValue(Capture(move)))<=BISHOP_BONUS &&
		(board.getMaterial(board.Side()).material_level() >= 7) &&
		mg.getPhase()<=Move_Generator::KILLER1_PHASE) {
                    node->extend += RECAPTURE_EXTENSION;
                    node->extensions |= RECAPTURE;
                    root->stats->recap_extensions++;
	    }
            if (options.search.repetition_extensions && rep_count == 1) {
	      node->extend += REPETITION_EXTENSION;
              node->extensions |= REPETITION;
              root->stats->rep_extensions++;
	    }
            if (options.search.threat_extensions && threat) {
                node->extend += THREAT_EXTENSION;
                node->extensions |= THREAT;
                root->stats->threat_extensions++;
            }
            if (limit) {
                node->extend /= 2;
            }
#ifdef RAZORING
            if (razor && node->extend == 0 && 
                (node->hibound==node->alpha+1) && 
                TypeOfMove(move) != Promotion &&
                (Capture(move) == EmptyPiece) && 
		board.getMaterial(board.Side()).material_level()>=3 &&
                !passed_pawn_push(board,move) &&
                in_check_after_move == NotInCheck) {
	        node->extend -= RAZOR_REDUCTION;
                root->stats->num_razored++;
#ifdef _TRACE
                if (master()) {
                    indent(ply); cout << "razoring, new depth = " << depth+RAZOR_REDUCTION << endl;
                }
#endif
            }
#endif
            else if (options.search.late_move_reductions && depth>=3*DEPTH_INCREMENT && 
               (node->hibound==node->alpha+1) && !in_check && 
                node->extend==0 &&
                CanHPrune(move) && 
                board.getMaterial(board.Side()).material_level()>3 &&
		!passed_pawn_push(board,move)) {
                  node->extend -= HISTORY_REDUCTION;
                  node->extensions |= HPRUNE;
            }
            if (node->extend > DEPTH_INCREMENT) node->extend = DEPTH_INCREMENT;
#ifdef FUTILITY_PRUNE
            int fmax;
            if (depth <= FUTILITY_RANGE && !in_check && 
                in_check_after_move == NotInCheck && 
		(node->hibound==node->alpha+1) && 
                node->extend <= 0 && 
                pruneOk(board,node->alpha,depth+node->extend,move,fmax,mat_score)) {
                int est = Gain(move);
                if (fmax + est < node->alpha) {
#ifdef _TRACE
                    if (master()) {
                         indent(ply); cout << "fwd pruned." << endl;
                    }
#endif
                    ++fpruned_moves;
                    continue;
                }
            }
#endif
#ifdef SMP
            // We are prepared to search a move, see if we can do so in 
            // another thread.
            // Note: We only split at low depths from the root, and we wait
            // until the remaining depth is high enough, because splitting
            // a shallow node is inefficient.
            if (!first && ply <= THREAD_SPLIT_LIMIT && 
 	        mg.getPhase() != Move_Generator::LAST_PHASE &&
                depth >= SPLIT_DEPTH[board.getMaterial(board.Side()).material_level()] &&
                !terminate) {
               // try to obtain an inactive thread
               ThreadInfo *ti;
               if ((ti = controller->pool->checkOut(this,node,ply,depth))!=NULL) {
                  // a thread is available, split off a new search,
                  // copying state so it can run independently:
#ifdef THREAD_TRACE
                  cout << "splitting" << endl;
#endif
	          ti->work->init(board,this,node,ti,ply);
                  // start searching in the new thread:
                  ti->start();
#ifdef SMP_STATS
                  root->stats->splits++;
#endif
                  continue;
               }
            }
#endif // SMP
            board.DoMove(move);
            if (!in_check && 
                board.any_attacks(board.KingPos(
                board.OppositeSide()),board.Side())) {
#ifdef _TRACE
                if (master()) {
                  indent(ply); cout << "Illegal move!" << endl;
                }
#endif
                board.UndoMove(move,node->state);
                continue;
            }
            // in rare cases we may not have applied the check extension
            // before the move, so do it here
            if (in_check_after_move == CheckUnknown) {
               in_check_after_move = board.CheckStatus();
               if (in_check_after_move == InCheck && node->extend == 0) {
                    node->extensions |= CHECK;
                    node->extend += CHECK_EXTENSION;
                    root->stats->check_extensions++;
                    if (limit) node->extend /= 2;
                    if (node->extend > DEPTH_INCREMENT) node->extend = DEPTH_INCREMENT;
               }
	    }
            ASSERT(in_check_after_move != CheckUnknown);
            ASSERT((in_check_after_move == InCheck) ==
                board.any_attacks(board.KingPos(
                board.Side()),board.OppositeSide()));
            if (depth+node->extend-DEPTH_INCREMENT >= 0) {
                try_score = -search(board, -node->hibound, -node->alpha,
                    ply+1, depth+node->extend-DEPTH_INCREMENT);
            }
            else {
              try_score = -quiesce(board, -node->hibound, -node->alpha,
                    ply+1, 0);
            }
            if (try_score != Illegal && !first && (try_score > node->alpha) && (try_score < node->beta) && !terminate) {
                // We failed to get a cutoff and must re-search
#ifdef _TRACE
                if (master()) {
                   indent(ply); cout << "score = " << try_score << " - no cutoff, researching .." << endl;
                   indent(ply); cout << "window = [" << node->alpha << "," << node->beta
                       << "]" << endl;
                }
#endif
                if (depth+node->extend-DEPTH_INCREMENT >= 0)
                    try_score=-search(board,
                        -node->beta,-node->alpha,ply+1,depth+node->extend-DEPTH_INCREMENT);
                else
                    try_score=-quiesce(board,-node->beta,-node->alpha,ply+1,0);
#ifdef _TRACE
                if (master()) {
                    indent(ply);
                    cout << ply << ". ";
                    MoveImage(move,cout);
                    cout << ' ' << try_score << endl;
                }
#endif
            }
            if ((node->extensions & HPRUNE) && try_score >= node->beta) {
              // verify LMR
              if (depth-DEPTH_INCREMENT >= 0) {
                  try_score = -search(board, -node->beta, -node->alpha,
                     ply+1, depth-DEPTH_INCREMENT);
               }
               else {
                   try_score = -quiesce(board, -node->beta, -node->alpha,
                      ply+1, 0);
               }
            }
            board.UndoMove(move,node->state);
            if (terminate) {
	      break;
            }
#ifdef _TRACE
            if (master()) {
                indent(ply);
                cout << ply << ". ";
                MoveImage(move,cout);
                cout << ' ' << try_score;
                if (first) cout << " (pv)";
                cout << endl;
            }
#endif
            if (try_score != Illegal) {
                first = 0;
                node->done[node->num_try++] = move;
                ASSERT(node->num_try<Constants::MaxMoves);
                if (try_score > node->alpha) {
		   node->best = move;
                   if (update_move(board,node,node,move,try_score,ply,depth)) {
		     // cutoff
		     break;
		   }
		}
                if (try_score >= Constants::BIG-1-ply) {
                   break;                        // mating move found
                }
                // zero-width window for the rest of the moves
                node->hibound = node->alpha+1;
            }
        }                                         // end move loop
#ifdef _TRACE
        if (node->score >= node->beta && master()) {
            indent(ply);
            cout << "**CUTOFF**" << endl;
        }
#endif
#ifdef SMP
    // We may have other threads still searching.
    if (node->split) {
       if (node->cutoff || terminate) {
          // We just need to stop the other threads.
          // We don't care about their results.
          stopAllThreads(node);
       }
       waitForThreadCompletion(node);
    }
#endif
    if (terminate) {
       node->score = node->alpha;
       goto search_end2;
    }
    if (node->num_try == 0
#ifdef FUTILITY_PRUNE
            && fpruned_moves == 0
#endif
        ) {
            // no moves were tried
            if (in_check) {
#ifdef _TRACE
                    if (master()) {
                       indent(ply); cout << "mate" << endl;
                    }
#endif
                    node->score = -(Constants::BIG - ply);
                    node->pv_length = 0;
                    goto search_end2;
            }
            else {                                // stalemate
#ifdef _TRACE
                if (master()) {
                   indent(ply); cout << "stalemate!" << endl;
                }
#endif
                node->score = draw_score(board);
                if (node->score > node->alpha && node->score < node->beta) {
		  node->pv[ply] = NullMove;
                  node->pv_length = 0;
		}
                return node->score; 
            }
        }
    }
    if (node->num_try && (initial_alpha != node->alpha)) {
		root->update_history(node,node->best,
                       node->cutoff,depth,
                       board.Side());
    }

    // don't insert into the hash table if we are terminating - we may
    // not have an accurate score.
 search_end:
    if (controller->use_hash_table && !terminate) {
        if (node->best == NullMove) node->best = hash_move;
        // store the position in the hash table, if there's room
        int value = node->score;
        Position_Info::ValueType val_type;
        // Adjust mate scores to reflect current ply. But only
        // if the score is in bounds.
        if (value > initial_alpha && value < node->beta) {
            if (value < -Constants::BIG+200) {
                value -= ply;
            }
            else if (value > Constants::BIG-200) {
                value += ply;
            }
        }
#ifdef _TRACE
        char typeChar;
#endif
        if (value <= initial_alpha) {
            val_type = Position_Info::UpperBound;
#ifdef _TRACE
            typeChar = 'U';
#endif
            // We don't have a "best" move, because all moves
            // caused alpha cutoff.  But if there was a hash
            // move or an initial move produced by internal
            // interative deepening, save it in the hash table
            // so it will be tried first next time around.
            node->best = hash_move;
        }
        else if (value >= node->beta) {
            val_type = Position_Info::LowerBound;
#ifdef _TRACE
            typeChar = 'L';
#endif
        }
        else {
#ifdef FUTILITY_PRUNE
            val_type = fpruned_moves ? Position_Info::Invalid : Position_Info::Valid;
#else
            val_type = Position_Info::Valid;
#endif
#ifdef _TRACE
            typeChar = 'E';
#endif
        }
#ifdef _TRACE
        if (master()) {
            indent(ply);
            cout << "storing type=" << typeChar <<
                " ply=" << ply << " depth=" << depth << " value=" << value <<
                " move=";
            MoveImage(node->best,cout); 
            cout << endl;
        }
#endif
        Hash_Entry new_pos(board.HashCode(rep_count), depth,
            controller->age,
            val_type, 
            board.CheckStatus() == InCheck,
            node->score, initial_alpha+1<node->beta,node->best);
        Lock(hash_lock);
        storeHash(new_pos,*(root->stats));
        Unlock(hash_lock);
    }
search_end2:
#ifdef MOVE_ORDER_STATS
    if (node->num_try && node->score != initial_alpha) {
       root->stats->move_order_count++;
       ASSERT(node->best_count>=0);
       if (node->best_count<4) {
           root->stats->move_order[node->best_count]++;
       }
    }
#endif
#ifdef FUTILITY_PRUNE
    root->stats->futility_pruning += fpruned_moves;
#endif
    int score = node->score;
    ASSERT(score >= -Constants::BIG && score <= Constants::BIG);
    return score;
}




