// Copyright 1994-2007 by Jon Dart.  All Rights Reserved.

#ifndef _SEARCH_H
#define _SEARCH_H

#include "board.h"
#include "movegen.h"
#include "hash.h"
#include "stats.h"
#include "pawnentr.h"
#include "arasset.h"
#include "searchc.h"
#ifdef SMP
#include "threadp.h"
#endif
#include <memory.h>
#include <time.h>
#include <set>
using namespace std;

#ifdef SMP
typedef 
  struct {
    ArasanSet<ThreadInfo *,Constants::MaxCPUs> slaves;
    int ply;
    int depth;
    int index; // in parent move list
  } split_t;
#endif

// Per-node info, part of search history stack
struct NodeInfo {
    NodeInfo() : cutoff(0),best(NullMove)
#ifdef SMP
      ,split(NULL) 
#endif
    {
      LockInit(nodeLock);
    }
#ifdef _WIN32
    virtual ~NodeInfo () {
      LockDestroy(nodeLock);
    }
#endif
    int score;
    int alpha, beta, hibound;
    int cutoff;
    int num_try;
    int flags; 
    Move best;
    Move last_move;
    int extend; // extension to apply
    int extensions; // mask of extensions
    Move pv[Constants::MaxPly];
    int pv_length;
    Move done[Constants::MaxMoves];
    int best_count;
    Board_State state;
#ifdef SMP
    LockDefine(nodeLock);
    // set of other threads executing searches that are siblings to this one.
    // (null if none)
    split_t *split;
#endif
};

struct QSearchHashEntry {
   hash_t hash_code;
   Move best;
   int score;
};

struct RootNodeInfo : public NodeInfo {
  int move_count;
  Move *moves;
};

class RootSearch;
class SearchController;

// There are 4 levels of verbosity.  Silent mode does no output to
// the console - it is used by the Windows GUI version.  Debug
// level is used to output debug info.  Whisper is used to "whisper"
// comments on a chess server.  "Trace" is used in trace mode
// (-t) of the chess server client.
enum Talk_Level { Silent, Debug, Whisper, Trace };
    
enum Search_Type { Fixed_Depth, Time_Limit, Fixed_Time };

typedef void (*PostFunction)(const Statistics &);
typedef int (*TerminateFunction)(const Statistics &);

class Search {
#ifdef SMP
    friend struct ThreadInfo;
    friend class ThreadPool;
#endif
    friend class Move_Ordering;

    // Encapsulates alpha-beta search routine. 

    enum Extension_Type { RECAPTURE=1, CHECK=2, PAWN_PUSH=4, CAPTURE=8,
       FORCED=16, REPETITION=32, HPRUNE=64, THREAT=128 };

    enum SearchFlags { IID=1, VERIFY=2 };

public:

    Search(SearchController *c);
        
    void init(const Board &board, Search *parent);

#ifdef SMP
    void init(const Board &board, Search *parent, NodeInfo *parentNode, ThreadInfo *ti, int ply);
#endif

    virtual ~Search();
    
    int is_draw( const Board &board, int &rep_count, int ply );
    
    int draw_score(const Board & board);
    
    int search(Board & board, int alpha, int beta,
               int ply, int depth, int flags = 0) {
      PUSH(board,alpha,beta,flags);
      return POP(search(board,ply,depth));
    }

    int search(Board &board, int ply, int depth);

    int master() const { return parent == this; }

#ifdef SMP
    // perform a subsidiary search in a separate thread
    void searchSMP(ThreadInfo *);
#endif

    int quiesce(Board & board, int alpha, int beta,
                int ply, int depth) {
       PUSHQ(board,alpha,beta);
       return POP(quiesce(board,ply,depth));
    }

    int quiesce(Board &board,int ply,int depth);

    int evalu8(const Board & board, int alpha, int beta, int ply);

    Move getBest() const;

    Search * split(const Board &,int ply);

    void terminateSearch() {
       terminate = 1;
    }

    NodeInfo *getParentNode() const {
      return parentNode;
    }
    
    int update_move(const Board &,
		    NodeInfo *parentNode, NodeInfo* myNode, Move move, 
                    int score, int ply, int depth);

    NodeInfo * getNode() const {
      return node;
    }

    static const int PAWN_HASH_SIZE = 8192;
    static const int KING_PAWN_HASH_SIZE = 65536;
    static const int KING_COVER_HASH_SIZE = 8192;
    static const int QSEARCH_CACHE_SIZE = 2*65536;

    struct King_Pawn_Entry {
        // defines one entry in the king/pawn hash table.
        hash_t hc;                                    // hashcode
        signed char white_king_position, black_king_position;
        signed char white_endgame_pawn_proximity;
        signed char black_endgame_pawn_proximity;
        byte w_uncatchable, b_uncatchable;
        byte pawns_both_sides;
        int score;
    };

    struct King_Cover_Entry {
       hash_t hc;
       int cover;
       int pawn_proximity;
       int king_out;
    };

    King_Pawn_Entry *king_pawn_hash_table;
    Pawn_Entry *pawn_hash_table;
    King_Cover_Entry *king_cover_hash_table;
    QSearchHashEntry *qsearch_cache;

    void clearHashTables();

protected:
    void *king_pawn_hash_table_allocated;
    void *score_hash_table_allocated;
    void *pawn_hash_table_allocated;
    void *king_cover_hash_table_allocated;
    void *qsearch_cache_allocated;
    Search *parent;
    NodeInfo *parentNode;
    SearchController *controller;
    RootSearch *root;
    Board board;
#ifdef SMP
    LockDefine(split_lock);
    split_t split_stack[SPLIT_STACK_MAX_DEPTH];
    split_t *split_stack_end;
    ThreadInfo *ti;
#endif
    int terminate;
    NodeInfo *node; // pointer into node stack, below
    NodeInfo stack[Constants::MaxPly];

    void update_pv(const Board &, const Move &m, int ply);

    void update_pv(const Board &,NodeInfo *node,NodeInfo *fromNode,Move move,int ply);

    void stopAllThreads();
#ifdef SMP
    void stopAllThreads(NodeInfo *);
    void waitForThreadCompletion(NodeInfo *);
#endif

private:

    // make this class uncopyable:
    Search( const Search & );
    Search &operator = ( const Search & );

    int pruneOk(const Board &board, int alpha,
		int depth,Move move, int &fmax, int &mat_score);

    int qsearch_check(Board &board,int ply,int depth);
    int qsearch_no_check(Board &board,int ply,int depth);

    FORCEINLINE void PUSH(Board &board,int alpha,int beta,int flags) {
       ++node; 
       node->alpha = node->score = alpha; 
       node->beta = node->hibound = beta; 
       node->flags = flags; 
       node->best = NullMove; 
       node->num_try = 0;
#ifdef SMP
       node->split = NULL;
#endif
       node->cutoff = 0; 
	}

    FORCEINLINE void PUSHQ(Board &board,int alpha,int beta) {
       ++node; 
       node->alpha = node->score = alpha; 
       node->beta = node->hibound = beta; 
       node->best = NullMove; 
	}

    FORCEINLINE int POP(int value)  {
       int val = value; 
       --node; 
       return val;
    }

};

class RootSearch : public Search {
    friend class Search;
    friend class SearchController;
public: 
    RootSearch(SearchController *c);

    virtual ~RootSearch() {
    }

    Move ply0_search(const Board &board, 
                Search_Type srcType,
                int32 timelimit,
                int32 xtratime,
                int plylimit,
                const int background,
                Statistics &stat_buf,
                Talk_Level t,
                Move *exclude, int num_exclude);

    time_t getTimeAdded() {
        return time_added;
    }

    // Get the starting time for the search in centiseconds
    time_t get_start_time() const
    {
        return start_time;
    }
    
    Search_Type get_search_type() const
    {
        return type_of_search;
    }
    
    void set_time_limit(time_t limit,time_t xtra_time_limit)
    {
       type_of_search = Time_Limit;
       time_target = limit;
       xtra_time = xtra_time_limit;
    }
    
    void set_ply_limit(int ply)
    {
       type_of_search = Fixed_Depth;
       ply_limit = ply;
    }
    
    time_t get_time_limit() const
    { 
        return time_target + (time_added ? xtra_time : 0);
    }
    
    int is_background_search() const
    {
        return bkgrnd_search;
    }
    
    int update_root_move(const Board &,NodeInfo *parentNode, NodeInfo *node, Move move, int score);

    void set_background(int b);

    void terminate_now();

    int was_terminated() const {
      return terminated;
    }

private:
    void init(const Board &);

    int check_time(const Board &,int ply, int rank, int interation_depth);
        
    // output something to the console
    void show_status(const Board &board,const Move bestMove,int faillow,int failhigh,int complete);

    void iterate();

    int ply0_search(Board &board, int alpha, int beta,
            int depth, Move exclude [], int num_exclude, int multipv_count);

    void update_stats(int alpha, int beta,int multipv_count);

    void update_history(NodeInfo *parentNode, Move move, int cutoff, 
       int depth, ColorType side);

    // Do initial ordering of moves returned by the move generator.
    // Set "easy" = true if the 1st (pv) move is measurably better than
    //the others.
    int order_ply0_moves(Board &board, const Move &best_move );	    

    // Re-order ply 0 moves after the initial sort
    int reorder_ply0_moves( const Board &board, const Move &best_move);

    int score(Board &board, const Move &move);

    Move excludes[MAX_PV];
    int num_excludes;
    Search_Type type_of_search;
    int easy;
    Move easy_move;
    Move ply0first; // first move to search
    Move ply0moves[Constants::MaxMoves];
    int ply0move_count;
    ColorType computerSide;
    ColorType side_to_move;
    Statistics *stats;
    time_t time_target, start_time, xtra_time, time_added;
    int time_check_counter;
    time_t last_time;
    int ply_limit;
    int bkgrnd_search;
    int root_total_mat;
    int null_depth;
    int terminated, time_up;
    Board initial_board;
    StateType state;
    int iteration_depth;
    int display_score; // The score displayed to the user
    SearchContext context;

};

class SearchController {
    // Manages shared state that is not particular to a search
    // thread
    friend class Search;
    friend class RootSearch;
    friend struct ThreadInfo;
public:

    SearchController();

    virtual ~SearchController();

    Move find_best_move( 
            const Board &board, 
            Search_Type srcType,
            int32 time_limit,
            int32 xtratime,
            int ply_limit,
            int background,
            int uci,
            Statistics &stats,
            Talk_Level t);
    
    Move find_best_move( 
            const Board &board, 
            Search_Type srcType,
            int32 time_limit,
            int32 xtratime,
            int ply_limit,
            int background,
            int uci,
            Statistics &stats,
            Talk_Level t,
            Move *exclude,
            int num_excludes);
    
    void register_command_function(void (*cmd)(char *,int &)) {
       command_function = cmd;
    }
     
    void register_pending_function(void (*cmd)(const char *)) {
       pending_function = cmd;
    }
     
    PostFunction register_post_function(PostFunction post) {
       PostFunction tmp = post_function;
       post_function = post;
       return tmp;
    }
     
    TerminateFunction register_terminate_function(TerminateFunction term) {
       TerminateFunction tmp = terminate_function;
       terminate_function = term;
       return tmp;
    }
     
    void set_background(int b);
    
    void init();

    void reinit();

    void setRatingDiff(int rdiff);

    // check console input
    int check_input(const Board &);

    // return # of nodes visited in current search
    uint64 get_numnodes() const;
            
    Talk_Level get_talk_level() const {
       return talkLevel;
    }

    void set_talk_level(Talk_Level t)
    {
       talkLevel = t;
    }
    
    int draw_score(const Board & board, ColorType computerSide,
       int root_total_mat);
    
    void set_time_limit(time_t limit,time_t xtra_time_limit) {
       rootSearch->set_time_limit(limit,xtra_time_limit);
    }

    RootSearch *getRootSearch() const {
      return rootSearch;
    }

    void clearHashTables();

private:
    void cleanup(int clearHash);

    // pointer to function, called to process a command
    // received during the search (console programs only).
    void (*command_function)(char *,int &);

    // pointer to function, called to output status during
    // a search (used in console mode only).
    void (*post_function)(const Statistics &);

    int (*terminate_function)(const Statistics &);

    // Function to add a command to the deferred execution
    // stack (UNIX_ONLY)
    void (*pending_function)(const char*);

    // Calculate the recommended hash table size (expressed
    // in number of entries).  
    static long calc_hash_size();
    
    RootSearch *rootSearch;
    int age;
    int explicit_excludes;
    int use_hash_table;
    int hash_table_size;
    int ratingDiff;
    int ratingFactor; 
    int pipe;
    int terminated;
    int uci;
    Talk_Level talkLevel;

    LockDefine(input_lock);

#ifdef SMP
    ThreadPool *pool;
#endif
};

#endif
